# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the different types of morphology patterns used
# by the basic morphology processor included with SimpleNLG.
class Inflection(Enum):
    # The morphology processor has simple rules for pluralising Greek and Latin
    # nouns. The full list can be found in the explanation of the morphology
    # processor.
    GRECO_LATIN_REGULAR = 0
    # A word having an irregular pattern essentially means that none of the
    # supplied rules can be used to correctly inflect the word. The inflection
    # should be defined by the user or appear in the lexicon. sheep is
    #an example of an irregular noun.
    IRREGULAR = 1
    # Regular patterns represent the default rules when dealing with
    # inflections. A full list can be found in the explanation of the
    # morphology processor. An example would be adding -s to the end
    # of regular nouns to pluralise them.
    REGULAR = 2
    # Regular double patterns apply to verbs where the last consonant is
    # duplicated before applying the new suffix. For example, the verb
    # tag has a regular double pattern as its inflected forms include
    # tagged and tagging.
    REGULAR_DOUBLE = 3
    # The value for uncountable nouns, which are not inflected in their plural
    # form.
    UNCOUNT = 4
    # The value for words which are invariant, that is, are never inflected.
    INVARIANT = 5

    # convenience method: parses an inflectional code such as
    # "irreg|woman|women" to retrieve the first element, which is the code
    # itself, then maps it to the value of Inflection.
    @staticmethod
    def getInflCode(code):
        code = code.lower().strip()
        infl = None
        if code == "reg":
            infl = Inflection.REGULAR
        elif code == "irreg":
            infl = Inflection.IRREGULAR
        elif code == "regd":
            infl = Inflection.REGULAR_DOUBLE
        elif code == "glreg":
            infl = Inflection.GRECO_LATIN_REGULAR
        elif code in ["uncount", "noncount", "groupuncount"]:
            infl = Inflection.UNCOUNT
        elif code == "inv":
            infl = Inflection.INVARIANT
        return infl

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the different types of interrogatives or
# questions that SimpleNLG can realise. T
class InterrogativeType(Enum):
    # The type of interrogative relating to the manner in which an event
    # happened. For example, John kissed Mary becomes
    # How did John kiss Mary?
    HOW = 0
    # A how question related to a predicative sentence, such as John is fine, which becomes How is John?
    HOW_PREDICATE = 1
    # This type of interrogative is a question pertaining to the object of a
    # phrase. For example, John bought a horse becomes what did
    # John buy? while John gave Mary a flower becomes
    # What did
    # John give Mary?
    WHAT_OBJECT = 2
    # This type of interrogative is a question pertaining to the subject of a
    # phrase. For example, A hurricane destroyed the house becomes
    # what destroyed the house?
    WHAT_SUBJECT = 3
    # This type of interrogative concerns the object of a verb that is to do
    # with location. For example, John went to the beach becomes
    # Where did John go?
    WHERE = 4
    # This type of interrogative is a question pertaining to the indirect
    # object of a phrase when the indirect object is a person. For example,
    # John gave Mary a flower becomes
    # Who did John give a flower to?
    WHO_INDIRECT_OBJECT = 5
    # This type of interrogative is a question pertaining to the object of a
    # phrase when the object is a person. For example,
    # John kissed Mary becomes who did John kiss?
    WHO_OBJECT = 6
    # This type of interrogative is a question pertaining to the subject of a
    # phrase when the subject is a person. For example,
    # John kissed Mary becomes Who kissed Mary? while
    # John gave Mary a flower becomes Who gave Mary a flower?
    WHO_SUBJECT = 7
    # The type of interrogative relating to the reason for an event happening.
    # For example, John kissed Mary becomes Why did John kiss
    # Mary?
    WHY = 8
    # This represents a simple yes/no questions. So taking the example phrases
    # of John is a professor and John kissed Mary we can
    # construct the questions Is John a professor? and
    # Did John kiss Mary?
    YES_NO = 9
    # This represents a "how many" questions. For example of
    # dogs chased John/em> becomes How many dogs chased John
    HOW_MANY = 10

    # A method to determine if the {@code InterrogativeType} is a question
    # concerning an element with the discourse function of an object.
    @classmethod
    def isObject(cls, otype):
        return otype in [cls.WHO_OBJECT, cls.WHAT_OBJECT]

    # A method to determine if the {@code InterrogativeType} is a question
    # concerning an element with the discourse function of an indirect object.
    @classmethod
    def isIndirectObject(cls, otype):
        return cls.WHO_INDIRECT_OBJECT==otype

    # Convenience method to return the String corresponding to the question
    # word. Useful, since the types in this enum aren't all simply convertible
    # to strings (e.g. WHO_SUBJCT and WHO_OBJECT both
    # correspond to String Who)
    def getString(self):
        otype = self
        if otype in [self.HOW, self.HOW_PREDICATE]:
            return "how"
        elif otype in [self.WHAT_OBJECT, self.WHAT_SUBJECT]:
            return "what"
        elif otype == self.WHERE:
            return "where"
        elif otype in [self.WHO_INDIRECT_OBJECT, self.WHO_OBJECT, self.WHO_SUBJECT]:
            return "who"
        elif otype==self.WHY:
            return "why"
        elif otype==self.HOW_MANY:
            return "how many"
        elif otype==self.YES_NO:
            return "yes/no"
        return ''

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from abc import ABC
from ..features.Feature             import *
from ..features.NumberAgreement     import *
from ..features.Tense               import *


# NLGElement is the base class that all elements extend from.
class NLGElement(ABC):
    def __init__(self):
        self.category    = None    # ElementCategory
        self.features    = {}      # {string, object}
        self.parent      = None    # NLGElement
        self.realisation = None    # string
        self.factory     = None    # NLGFactory

    # Sets the category of this element.
    def setCategory(self, new_category):
        self.category = new_category

    # Retrieves the category for this element.
    def getCategory(self):
        return self.category

    # Adds a feature to the feature map.
    def setFeature(self, featureName, featureValue):
        if not featureName:
            return
        if featureValue is None:
            if featureName in self.features:
                del self.features[featureName]
        else:
            self.features[featureName] = featureValue

    # Retrieves the value of the feature.
    def getFeature(self, featureName):
        return self.features.get(featureName, None)

    # Retrieves the value of the feature as a string. If the feature doesn't
    # exist then null is returned.
    def getFeatureAsString(self, featureName):
        return self.getFeature(featureName)

    # Retrieves the value of the feature as a list of elements.
    def getFeatureAsElementList(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, NLGElement):
            return [value]
        elif isinstance(value, list):
            elist = []
            for item in value:
                if isinstance(item, NLGElement):
                    elist.append(item)
            return elist
        return []

    # Retrieves the value of the feature as a list of java objects. If the feature
    def getFeatureAsList(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, list):
            return value
        return [value]

    # Retrieves the value of the feature as a list of strings.
    def getFeatureAsStringList(self, featureName):
        value = self.getFeature(featureName)
        if value is None:
            return []
        elif isinstance(value, list):
            values = []
            for item in value:
                values.append(str(item))
            return values
        return [str(value)]

    # Retrieves the value of the feature as an Integer.
    def getFeatureAsInteger(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, int):
            return value
        elif isinstance(value, float):
            return int(value)
        elif isinstance(value, str):
            try:
                val = int(str)
                return val
            except ValueError:
                return None

    # Retrieves the value of the feature as a Long.
    def getFeatureAsLong(self, featureName):
        return self.getFeatureAsInteger(featureName)

    # Retrieves the value of the feature as a Float.
    def getFeatureAsFloat(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, float):
            return value
        elif isinstance(value, int):
            return float(value)
        elif isinstance(value, str):
            try:
                val = float(str)
                return val
            except ValueError:
                return None

    # Retrieves the value of the feature as a Double.
    def getFeatureAsDouble(self, featureName):
        return self.getFeatureAsFloat(featureName)

    # Retrieves the value of the feature as a Boolean.
    def getFeatureAsBoolean(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, bool):
            return value
        return False

    # Retrieves the value of the feature as a NLGElement.
    def getFeatureAsElement(self, featureName):
        value = self.getFeature(featureName)
        if isinstance(value, NLGElement):
            return value
        elif isinstance(value, str):
            from .StringElement import StringElement    #prevent circular import
            return StringElement(value)
        return None

    # Retrieves the map containing all the features for this element.
    def getAllFeatures(self):
        return self.features

    # Checks the feature map to see if the named feature is present in the map.
    def hasFeature(self, featureName):
        return featureName in self.features

    # Deletes the named feature from the map.
    def removeFeature(self, featureName):
        if featureName in self.features:
            del self.features[featureName]

    # Deletes all the features in the map.
    def clearAllFeatures(self):
        self.features = {}

    # Sets the parent element of this element.
    def setParent(self, new_parent):
        self.parent = new_parent

    # Retrieves the parent of this element.
    def getParent(self):
        return self.parent

    # Sets the realisation of this element.
    def setRealisation(self, realised):
        self.realisation = realised

    # Retrieves the final realisation of this element.
    def getRealisation(self):
        start = 0
        end   = 0
        if self.realisation is not None:
            end = len(self.realisation)
            while start<len(self.realisation) and ' '==self.realisation[start]:
                start += 1
            if start == len(self.realisation):
                self.realisation = None
            else:
                while end>0 and ' '==self.realisation[end - 1]:
                    end -= 1
        # AG: changed this to return the empty string if the realisation is null
        # avoids spurious nulls appearing in output for empty phrases
        if self.realisation is None:
            return ''
        return self.realisation[start:end]

    def __str__(self):
        buffer = '{realisation='
        if self.realisation is not None:
            buffer += '"' + self.realisation + '"'
        else:
            buffer += '""'
        if self.category is not None:
            buffer += ", category=" + str(self.category)
        if self.features is not None:
            buffer += ", features={"
            for key, value in self.features.items():
                buffer += '%s:%s, ' % (key, str(value))
            buffer = buffer[:-2]
            buffer += '}'
        buffer += '}'
        return buffer

    # This makes the xxPhraseSpec(s) print correctly.  They inherit from PhraseSpec
    # which inherits from NLGElement.
    def __repr__(self):
        return self.__str__()

    def isA(self, checkCategory):
        if self.category is not None:
            return self.category.equalTo(checkCategory)
        elif checkCategory is None:
            return True

    # Retrieves the children for this element. This method needs to be
    # overridden for each specific type of element. Each type of element will
    # have its own way of determining the child elements.
    # return list of NGElement
    def getChildren(self):
        assert False, 'Not implemented in base.'

    # Retrieves the set of features currently contained in the feature map.
    def getAllFeatureNames(self):
        return self.features.keys()

    def printTree(self, indent=None):
        if indent is None:
            indent = ''
        this_indent  = indent + " |-"
        child_indent = indent + " |-"
        pstr = "NLGElement: " + str(self) + '\n'
        for child in self.getChildren():
            pstr += this_indent + child.printTree(child_indent)
        return pstr

    # Sets the number agreement on this element.
    def setPlural(self, isPlural):
        if isPlural:
            self.setFeature(Feature.NUMBER, NumberAgreement.PLURAL)
        else:
            self.setFeature(Feature.NUMBER, NumberAgreement.SINGULAR)

    # Determines if this element is to be treated as a plural. This is a
    # convenience method and not all element types make use of number
    # agreement.
    def isPlural(self):
        return NumberAgreement.PLURAL == self.getFeature(Feature.NUMBER)

    # Retrieves the tense for this element.
    def getTense(self):   # deprecated
        value = self.getFeature(Feature.TENSE)
        if isinstance(value, Tense):
            return value
        return Tense.PRESENT

    # Sets the tense on this element.
    def setTense(self, new_tense):
        self.setFeature(Feature.TENSE, new_tense)

    # Sets the negation on this element.
    def setNegated(self, is_negated):      # deprecated
        self.setFeature(Feature.NEGATED, is_negated)

    # Determines if this element is to be treated as a negation.
    def isNegated(self):    # deprecated
        return self.getFeature(Feature.NEGATED)

    # @return the NLG factory
    def getFactory(self):
        return self.factory

    # @param factory
    def setFactory(self, factory):
        self.factory = factory

    # An NLG element is equal to some object if the object is an NLGElement,
    # they have the same category and the same features.
    def __eq__(self, obj):
        if isinstance(obj, str):
            if obj is None and self.realisation is None:
                return True
            else:
                return obj == self.realisation
        elif isinstance(obj, NLGElement):
            return self.category == obj.category and self.features == obj.features
        return False

    def __ne__(self, obj):
        return not self.__eq__(obj)

    def __hash__(self):
        objects = []
        if self.features:
            objects.extend(self.features)
        if self.category:
            objects.append(self.category)
        return hash(tuple(objects))



from . ClauseStatus         import *
from . DiscourseFunction    import *
from . Feature              import *
from . Form                 import *
from . Gender               import *
from . Inflection           import *
from . InternalFeature      import *
from . InterrogativeType    import *
from . LexicalFeature       import *
from . NumberAgreement      import *
from . Person               import *
from . Tense                import *



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...framework.CoordinatedPhraseElement   import *
from ...framework.DocumentCategory           import *
from ...framework.DocumentElement            import *
from ...framework.ListElement                import *
from ...framework.NLGElement                 import *
from ...framework.NLGModule                  import *
from ...framework.StringElement              import *
from .NumberedPrefix                         import *

# This processing module adds some simple plain text formatting to the
# SimpleNLG output.
class TextFormatter(NLGModule):
    def __init__(self):
        super().__init__()
        self.numberedPrefix = NumberedPrefix()

    # @Override
    def initialise(self):
        pass # Do nothing

    # @Override
    def realise(self, element):
        if isinstance(element, NLGElement):
            return self._realiseElement(element)
        elif isinstance(element, list):
            return [self._realiseElement(eachElement) for eachElement in element]
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    # @Override
    def _realiseElement(self, element):
        realisation = ''
        category = element.getCategory()
        components = element.getChildren()
        #NB: The order of the if-statements below is important!
        # check if this is a canned text first
        if isinstance(element, StringElement):
            realisation += element.getRealisation()
        elif isinstance(category, DocumentCategory):
            title = None
            if isinstance(element, DocumentElement):
                title = element.getTitle()
            if category == DocumentCategory.DOCUMENT:
                realisation = self.appendTitle(realisation, title, 2)
                realisation = self.realiseSubComponents(realisation, components)
            elif category == DocumentCategory.SECTION:
                realisation = self.appendTitle(realisation, title, 1)
                realisation = self.realiseSubComponents(realisation, components)
            elif category == DocumentCategory.LIST:
                realisation = self.realiseSubComponents(realisation, components)
            elif category == DocumentCategory.ENUMERATED_LIST:
                self.numberedPrefix.upALevel()
                if title is not None:
                    realisation += title + '\n'
                if components:
                    realisedComponent = self.realise(components[0])
                    if realisedComponent is not None:
                        realisation += realisedComponent.getRealisation()
                    for i, component in enumerate(components):
                        if i==0: continue
                        if realisedComponent is not None and not realisedComponent.getRealisation().endswith("\n"):
                            realisation += ' '
                        if component.getParent().getCategory() == DocumentCategory.ENUMERATED_LIST:
                            self.numberedPrefix.increment()
                        realisedComponent = self.realise(component)
                        if realisedComponent is not None:
                            realisation += realisedComponent.getRealisation()
                self.numberedPrefix.downALevel()
            elif category == DocumentCategory.PARAGRAPH:
                if components:
                    realisedComponent = self.realise(components[0])
                    if realisedComponent is not None:
                        realisation += realisedComponent.getRealisation()
                    for i, component in enumerate(components):
                        if i==0: continue
                        if realisedComponent is not None:
                            realisation += ' '
                        realisedComponent = self.realise(component)
                        if realisedComponent is not None:
                            realisation += realisedComponent.getRealisation()
                realisation += "\n\n"
            elif category == DocumentCategory.SENTENCE:
                realisation += element.getRealisation()
            elif category == DocumentCategory.LIST_ITEM:
                if element.getParent() is not None:
                    if element.getParent().getCategory() == DocumentCategory.LIST:
                        realisation += " * "
                    elif element.getParent().getCategory() == DocumentCategory.ENUMERATED_LIST:
                        realisation += self.numberedPrefix.getPrefix() + " - "
                for eachComponent in components:
                    realisedComponent = self.realise(eachComponent)
                    if realisedComponent is not None:
                        realisation += realisedComponent.getRealisation()
                        if components.index(eachComponent) < len(components)-1:
                            realisation += ' '
                #finally, append newline
                realisation += "\n"
            # also need to check if element is a ListElement (items can
            # have embedded lists post-orthography) or a coordinate
        elif isinstance(element, ListElement) or isinstance(element, CoordinatedPhraseElement):
            for eachComponent in components:
                realisedComponent = self.realise(eachComponent)
                if realisedComponent is not None:
                    realisation += realisedComponent.getRealisation() + ' '
        return StringElement(realisation)

    # realiseSubComponents -- Realises subcomponents iteratively.
    def realiseSubComponents(self, realisation, components):
        for eachComponent in components:
            realisedComponent = self.realise(eachComponent)
            if realisedComponent is not None:
                realisation += realisedComponent.getRealisation()
        return realisation

    # appendTitle -- Appends document or section title to the realised document.
    def appendTitle(self, realisation, title, numberOfLineBreaksAfterTitle):
        if title:
            realisation += title
            for _ in range(numberOfLineBreaksAfterTitle):
                realisation += "\n"
        return realisation



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

import os
import logging
from   copy         import deepcopy
from   collections  import defaultdict
import xml.etree.ElementTree as ET
from   .Lexicon                     import *
from   ..features.Inflection        import *
from   ..features.LexicalFeature    import *
from   ..framework.LexicalCategory  import *
from   ..framework.WordElement      import *
from simplenlg import resource_directory

# This class loads words from an XML lexicon. All features specified in the
# lexicon are loaded
class XMLLexicon(Lexicon):
    XML_BASE     = "base"       # base form of Word
    XML_CATEGORY = "category"   # base form of Word
    XML_ID       = "id"         # base form of Word
    XML_WORD     = "word"       # node defining a word
    def __init__(self, lexicon_fn=None):
        super().__init__()
        if not lexicon_fn:
            lexicon_fn = os.path.join(resource_directory, 'default-lexicon.xml')
        self.words          = set()                    # all words in the lexicon
        self.indexByID      = {}                       # ID's are unique mapping to word
        self.indexByBase    = defaultdict(list)        # list of WordElements by base_form
        self.indexByVariant = defaultdict(list)        # list of WordElements by variants
        self.createLexicon(lexicon_fn)

    # Lexicon Note: not all entries have an ID
    # 'base' is the word but this is not unique either (multiple pos)
    def createLexicon(self, xmlfile):
        # Load the XML file to a list of dictionaries
        tree = ET.parse(xmlfile)
        for word in tree.getroot():
            if word.tag != self.XML_WORD :
                logging.warning('Unrecognized element: %s', word.tag)
                continue
            definition = {}
            for item in word:
                if len(item) != 0:
                    raise ValueError('Parsing error. There should be no sub-items')
                definition[item.tag] = item.text
            we = self.convertToWord(definition)
            self.words.add(we)
            self.IndexWord(we)
        self.addSpecialCases()

    # create a simplenlg WordElement from a Word node in a lexicon XML file
    def convertToWord(self, word_dict):
        word = WordElement()
        inflections = []
        for feature, value in word_dict.items():
            feature = feature.strip()
            if isinstance(value, str):
                value = value.strip()
            if not feature:
                logging.error("Error in XML lexicon for %s", word)
                break
            if feature.lower() == self.XML_BASE:
                word.setBaseForm(value)
            elif feature.lower() == self.XML_CATEGORY:
                word.setCategory(LexicalCategory[value.upper()])
            elif feature.lower() == self.XML_ID:
                word.setId(value)
            elif value is None or value=='':
                # if this is an infl code, add it to inflections
                infl = Inflection.getInflCode(feature)
                if infl is not None:
                    inflections.append(infl)
                # otherwise assume it's a boolean feature
                else:
                    word.setFeature(feature, True)
            else:
                word.setFeature(feature, value)
        # if no infl specified, assume regular
        if not inflections:
            inflections.append(Inflection.REGULAR)
        # default inflection code is "reg" if we have it, else random pick form infl codes available
        if Inflection.REGULAR in inflections:
            defaultInfl = Inflection.REGULAR
        else:
            defaultInfl = inflections[0]
        word.setFeature(LexicalFeature.DEFAULT_INFL, defaultInfl)
        word.setDefaultInflectionalVariant(defaultInfl)
        for infl in inflections:
            word.addInflectionalVariant(infl)
        return word

    def IndexWord(self, word):
        base_form = word.getBaseForm()
        if base_form:
            self.indexByBase[base_form].append(word)
        wid = word.getId()
        if wid:
            if wid in self.indexByID:
                logging.error("Lexicon error: ID %s occurs more than once", wid)
            else:
                self.indexByID[wid] = word
        for variant in self.getVariants(word):
            self.indexByVariant[variant].append(word)

    def addSpecialCases(self):
        be = self.getWord("be", LexicalCategory.VERB)
        if be is not None:
            self.indexByVariant['is'  ].append(be)
            self.indexByVariant['am'  ].append(be)
            self.indexByVariant['are' ].append(be)
            self.indexByVariant['was' ].append(be)
            self.indexByVariant['were'].append(be)

    # Override Lexicon see simplenlg.lexicon.Lexicon#getWords
    def getWords(self, baseForm, category):
        return self.getWordsFromIndex(baseForm, category, self.indexByBase)

    # get matching keys from an index map
    def getWordsFromIndex(self, indexKey, category, indexMap):
        results = []
        # case 1: unknown, return empty list
        if indexKey not in indexMap:
            return results
        # case 2: category is ANY, return everything
        if category == LexicalCategory.ANY:
            for word in indexMap[indexKey]:
                results.append(deepcopy(word))
            return results
        else:
            # case 3: other category, search for match
            for word in indexMap[indexKey]:
                if word.getCategory() == category:
                    results.append(deepcopy(word))
            return results

    # Override Lexicon see simplenlg.lexicon.Lexicon#getWordsByID
    def getWordsByID(self, wid):
        results = []
        if wid in self.indexByID:
            results.append(deepcopy(self.indexByID[wid]))
        return results

    # Override Lexicon see simplenlg.lexicon.Lexicon#getWordsFromVariant
    def getWordsFromVariant(self, variant, category):
        return self.getWordsFromIndex(variant, category, self.indexByVariant)

    # quick-and-dirty routine for getting morph variants should be replaced by something better!
    def getVariants(self, word):
        variants = set()
        variants.add(word.getBaseForm())
        category = word.getCategory()
        if isinstance(category, LexicalCategory):
            if category == LexicalCategory.NOUN:
                variants.add(self.getVariant(word, LexicalFeature.PLURAL, "s"))
            elif category == LexicalCategory.ADJECTIVE:
                variants.add(self.getVariant(word, LexicalFeature.COMPARATIVE, "er"))
                variants.add(self.getVariant(word, LexicalFeature.SUPERLATIVE, "est"))
            elif category == LexicalCategory.VERB:
                variants.add(self.getVariant(word, LexicalFeature.PRESENT3S, "s"))
                variants.add(self.getVariant(word, LexicalFeature.PAST, "ed"))
                variants.add(self.getVariant(word, LexicalFeature.PAST_PARTICIPLE, "ed"))
                variants.add(self.getVariant(word, LexicalFeature.PRESENT_PARTICIPLE, "ing"))
        return variants

    # quick-and-dirty routine for computing morph forms Should be replaced by something better!
    def getVariant(self, word, feature, suffix):
        if word.hasFeature(feature):
            return word.getFeatureAsString(feature)
        else:
            return self.getForm(word.getBaseForm(), suffix)

    # quick-and-dirty routine for standard orthographic changes Should be replaced by something better!
    def getForm(self, base, suffix):
        # add a suffix to a base form, with orthographic changes
        # rule 1 - convert final "y" to "ie" if suffix does not start with "i"
        # eg, cry + s = cries , not crys
        if base.endswith("y") and not suffix.startswith("i"):
            base = base[0:-1] + "ie"
        # rule 2 - drop final "e" if suffix starts with "e" or "i"
        # eg, like+ed = liked, not likeed
        if base.endswith("e") and (suffix.startswith("e") or suffix.startswith("i")):
            base = base[0:-1]
        # rule 3 - insert "e" if suffix is "s" and base ends in s, x, z, ch, sh
        # eg, watch+s -> watches, not watchs
        if suffix.startswith("s") and (base.endswith("s") or base.endswith("x") \
                or base.endswith("z") or base.endswith("ch") or base.endswith("sh")):
            base = base + "e"
        # have made changes, now append and return
        return base + suffix # eg, want + s = wants



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..features.ClauseStatus                import *
from ..features.Feature                     import *
from ..features.InternalFeature             import *
from ..features.LexicalFeature              import *
from ..framework.CoordinatedPhraseElement   import *
from ..framework.InflectedWordElement       import *
from ..framework.LexicalCategory            import *
from ..framework.NLGElement                 import *
from ..framework.PhraseCategory             import *
from ..framework.PhraseElement              import *
from ..framework.WordElement                import *
from .VPPhraseSpec                          import *
from .AdvPhraseSpec                         import *


# This class defines a clause (sentence-like phrase).
class SPhraseSpec(PhraseElement):
    vpFeatures = [Feature.MODAL, Feature.TENSE, Feature.NEGATED, Feature.NUMBER, Feature.PASSIVE, \
                  Feature.PERFECT, Feature.PARTICLE, Feature.PERSON, Feature.PROGRESSIVE, \
                  InternalFeature.REALISE_AUXILIARY, Feature.FORM, Feature.INTERROGATIVE_TYPE]

    def __init__(self, phraseFactory):
        super().__init__(PhraseCategory.CLAUSE)
        self.setFactory(phraseFactory)
        # create VP
        self.setVerbPhrase(phraseFactory.createVerbPhrase())
        # set default values
        self.setFeature(Feature.ELIDED, False)
        self.setFeature(InternalFeature.CLAUSE_STATUS, ClauseStatus.MATRIX)
        self.setFeature(Feature.SUPRESSED_COMPLEMENTISER, False)
        self.setFeature(LexicalFeature.EXPLETIVE_SUBJECT, False)
        self.setFeature(Feature.COMPLEMENTISER, phraseFactory.createWord( \
                "that", LexicalCategory.COMPLEMENTISER))

    # intercept and override setFeature, to set VP features as needed
    # adds a feature, possibly to the underlying VP as well as the SPhraseSpec itself
    # @Override
    def setFeature(self, featureName, featureValue):
        super().setFeature(featureName, featureValue)
        if featureName in self.vpFeatures:
            verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
            if isinstance(verbPhrase, VPPhraseSpec):
                verbPhrase.setFeature(featureName, featureValue)

    # adds a premodifier, if possible to the underlying VP
    # @Override
    def addPreModifier(self, newPreModifier):
        verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        if verbPhrase is not None:
            if isinstance(verbPhrase, PhraseElement):
                verbPhrase.addPreModifier(newPreModifier)
            elif isinstance(verbPhrase, CoordinatedPhraseElement):
                verbPhrase.addPreModifier(newPreModifier)
            else:
                super().addPreModifier(newPreModifier)

    # adds a complement, if possible to the underlying VP
    # @Override
    def addComplement(self, newComplement):
        verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        if isinstance(verbPhrase, VPPhraseSpec):
            verbPhrase.addComplement(newComplement)
        else:
            super().addComplement(newComplement)

    # @see simplenlg.framework.NLGElement#getFeature(java.lang.String)
    # @Override
    def getFeature(self, featureName):
        if super().getFeature(featureName) is not None:
            return super().getFeature(featureName)
        if  featureName in self.vpFeatures:
            verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
            if isinstance(verbPhrase, VPPhraseSpec):
                return verbPhrase.getFeature(featureName)
        return None

    # @return VP for this clause
    def getVerbPhrase(self):
        return self.getFeatureAsElement(InternalFeature.VERB_PHRASE)

    def setVerbPhrase(self, vp):
        self.setFeature(InternalFeature.VERB_PHRASE, vp)
        vp.setParent(self) # needed for syntactic processing

    # Set the verb of a clause
    def setVerb(self, verb):
        # get verb phrase element (create if necessary)
        verbPhraseElement = self.getVerbPhrase()
        # set head of VP to verb (if this is VPPhraseSpec, and not a coord)
        if isinstance(verbPhraseElement, VPPhraseSpec):
            verbPhraseElement.setVerb(verb)

    # Returns the verb of a clause
    def getVerb(self):
        # WARNING - I don't understand verb phrase, so this may not work!!
        verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        if isinstance(verbPhrase, VPPhraseSpec):
            return verbPhrase.getHead()
        else:
            # return null if VP is coordinated phrase
            return None

    # Sets the subject of a clause (assumes this is the only subject)
    def setSubject(self, subject):
        if isinstance(subject, PhraseElement) or isinstance(subject, CoordinatedPhraseElement):
            subjectPhrase = subject
        else:
            subjectPhrase = self.getFactory().createNounPhrase(subject)
        subjects = [subjectPhrase]
        self.setFeature(InternalFeature.SUBJECTS, subjects)

    # Returns the subject of a clause (assumes there is only one)
    def getSubject(self):
        subjects = self.getFeatureAsElementList(InternalFeature.SUBJECTS)
        if not subjects:
            return None
        return subjects[0]

    # Sets the direct object of a clause (assumes this is the only direct object)
    def setObject(self, obj):
        # get verb phrase element (create if necessary)
        verbPhraseElement = self.getVerbPhrase()
        # set object of VP to verb (if this is VPPhraseSpec, and not a coord)
        if isinstance(verbPhraseElement, VPPhraseSpec):
            verbPhraseElement.setObject(obj)

    # Returns the direct object of a clause (assumes there is only one)
    def getObject(self):
        verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        if isinstance(verbPhrase, VPPhraseSpec):
            return verbPhrase.getObject()
        else:
            # return null if VP is coordinated phrase
            return None

    # Set the indirect object of a clause (assumes this is the only direct
    # indirect object)
    def setIndirectObject(self, indirectObject):
        # get verb phrase element (create if necessary)
        verbPhraseElement = self.getVerbPhrase()
        # set head of VP to verb (if this is VPPhraseSpec, and not a coord)
        if isinstance(verbPhraseElement, VPPhraseSpec):
            verbPhraseElement.setIndirectObject(indirectObject)

    # Returns the indirect object of a clause (assumes there is only one)
    def getIndirectObject(self):
        verbPhrase = self.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        if isinstance(verbPhrase, VPPhraseSpec):
            return verbPhrase.getIndirectObject()
        else:
            # return null if VP is coordinated phrase
            return None

    # Add a modifier to a clause Use heuristics to decide where it goes
    # @Override
    def addModifier(self, modifier):
        # adverb is frontModifier if sentenceModifier
        # otherwise adverb is preModifier
        # string which is one lexicographic word is looked up in lexicon,
        # above rules apply if adverb
        # Everything else is postModifier
        if modifier is None:
            return
        # get modifier as NLGElement if possible
        modifierElement = None
        if isinstance(modifier, NLGElement):
            modifierElement = modifier
        elif isinstance(modifier, str):
            modifierString = modifier
            if len(modifierString)>0 and not " " in modifierString:
                modifierElement = self.getFactory().createWord(modifier, LexicalCategory.ANY)
        # if no modifier element, must be a complex string
        if modifierElement is None:
            self.addPostModifier(str(modifier))
            return
        # AdvP is premodifer (probably should look at head to see if
        # sentenceModifier)
        if isinstance(modifierElement, AdvPhraseSpec):
            self.addPreModifier(modifierElement)
            return
        # extract WordElement if modifier is a single word
        if isinstance(modifierElement, WordElement):
            modifierWord = modifierElement
        elif isinstance(modifierElement, InflectedWordElement):
            modifierWord = modifierElement.getBaseWord()
        if modifierWord is not None and modifierWord.getCategory() == LexicalCategory.ADVERB:
            # adverb rules
            if modifierWord.getFeatureAsBoolean(LexicalFeature.SENTENCE_MODIFIER):
                self.addFrontModifier(modifierWord)
            else:
                self.addPreModifier(modifierWord)
            return
        # default case
        self.addPostModifier(modifierElement)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.ClauseStatus               import *
from ...features.DiscourseFunction          import *
from ...features.Feature                    import *
from ...features.Form                       import *
from ...features.InternalFeature            import *
from ...features.InterrogativeType          import *
from ...features.NumberAgreement            import *
from ...features.Person                     import *
from ...features.Tense                      import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.LexicalCategory           import *
from ...framework.ListElement               import *
from ...framework.PhraseCategory            import *
from ...framework.PhraseElement             import *
from ...phrasespec.SPhraseSpec              import *
from ...phrasespec.VPPhraseSpec             import *
from .PhraseHelper                          import *
from .VerbPhraseHelper                      import *

# This is a helper class containing the main methods for realising the syntax
# of clauses. It is used exclusively by the SyntaxProcessor.
class ClauseHelper(object):

    # The main method for controlling the syntax realisation of clauses.
    @classmethod
    def realise(cls, parent, phrase):
        realisedElement = None
        phraseFactory = phrase.getFactory()
        splitVerb = None
        interrogObj = False
        if phrase is not None:
            realisedElement = ListElement()
            verbElement = phrase.getFeatureAsElement(InternalFeature.VERB_PHRASE)
            if verbElement is None:
                verbElement = phrase.getHead()
            cls.checkSubjectNumberPerson(phrase, verbElement)
            cls.checkDiscourseFunction(phrase)
            cls.copyFrontModifiers(phrase, verbElement)
            cls.addComplementiser(phrase, parent, realisedElement)
            cls.addCuePhrase(phrase, parent, realisedElement)
            if phrase.hasFeature(Feature.INTERROGATIVE_TYPE):
                inter = phrase.getFeature(Feature.INTERROGATIVE_TYPE)
                interrogObj = inter in [InterrogativeType.WHAT_OBJECT, InterrogativeType.WHO_OBJECT, \
                                        InterrogativeType.HOW_PREDICATE, InterrogativeType.HOW, \
                                        InterrogativeType.WHY, InterrogativeType.WHERE]
                splitVerb = cls.realiseInterrogative(phrase, parent, realisedElement, phraseFactory, verbElement)
            else:
                PhraseHelper.realiseList(parent, realisedElement, \
                                         phrase.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS), \
                                         DiscourseFunction.FRONT_MODIFIER)
            cls.addSubjectsToFront(phrase, parent, realisedElement, splitVerb)
            passiveSplitVerb =cls.addPassiveComplementsNumberPerson(phrase, parent, realisedElement, verbElement)
            if passiveSplitVerb is not None:
                splitVerb = passiveSplitVerb
            # realise verb needs to know if clause is object interrogative
            cls.realiseVerb(phrase, parent, realisedElement, splitVerb, verbElement, interrogObj)
            cls.addPassiveSubjects(phrase, parent, realisedElement, phraseFactory)
            cls.addInterrogativeFrontModifiers(phrase, parent, realisedElement)
            cls.addEndingTo(phrase, parent, realisedElement, phraseFactory)
        return realisedElement

    # Adds to to the end of interrogatives concerning indirect
    # objects.
    @classmethod
    def addEndingTo(cls, phrase, parent, realisedElement, phraseFactory):
        if InterrogativeType.WHO_INDIRECT_OBJECT == phrase.getFeature(Feature.INTERROGATIVE_TYPE):
            word = phraseFactory.createWord("to", LexicalCategory.PREPOSITION)
            realisedElement.addComponent(parent.realise(word))

    # Adds the front modifiers to the end of the clause when dealing with
    # interrogatives.
    @classmethod
    def addInterrogativeFrontModifiers(cls, phrase, parent, realisedElement):
        currentElement = None
        if phrase.hasFeature(Feature.INTERROGATIVE_TYPE):
            for subject in phrase.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS):
                currentElement = parent.realise(subject)
                if currentElement is not None:
                    currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.FRONT_MODIFIER)
                    realisedElement.addComponent(currentElement)

    # Realises the subjects of a passive clause.
    @classmethod
    def addPassiveSubjects(cls, phrase, parent, realisedElement, phraseFactory):
        currentElement = None
        if phrase.getFeatureAsBoolean(Feature.PASSIVE):
            allSubjects = phrase.getFeatureAsElementList(InternalFeature.SUBJECTS)
            if allSubjects or phrase.hasFeature(Feature.INTERROGATIVE_TYPE):
                realisedElement.addComponent(parent.realise(phraseFactory.createPrepositionPhrase("by")))
            for subject in allSubjects:
                subject.setFeature(Feature.PASSIVE, True)
                if subject.isA(PhraseCategory.NOUN_PHRASE) or isinstance(subject, CoordinatedPhraseElement):
                    currentElement = parent.realise(subject)
                    if currentElement is not None:
                        currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SUBJECT)
                        realisedElement.addComponent(currentElement)

    # Realises the verb part of the clause.
    @classmethod
    def realiseVerb(cls, phrase, parent, realisedElement, splitVerb, verbElement, whObj):
        cls.setVerbFeatures(phrase, verbElement)
        currentElement = parent.realise(verbElement)
        if currentElement is not None:
            if splitVerb is None:
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.VERB_PHRASE)
                realisedElement.addComponent(currentElement)
            else:
                if isinstance(currentElement, ListElement):
                    children = currentElement.getChildren()
                    currentElement = children[0]
                    currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.VERB_PHRASE)
                    realisedElement.addComponent(currentElement)
                    realisedElement.addComponent(splitVerb)
                    for eachChild in range(1,len(children)):
                        currentElement = children[eachChild]
                        currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.VERB_PHRASE)
                        realisedElement.addComponent(currentElement)
                else:
                    currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.VERB_PHRASE)
                    if whObj:
                        realisedElement.addComponent(currentElement)
                        realisedElement.addComponent(splitVerb)
                    else:
                        realisedElement.addComponent(splitVerb)
                        realisedElement.addComponent(currentElement)

    @classmethod
    def setVerbFeatures(cls, phrase, verbElement):
        pass

    # Realises the complements of passive clauses also sets number, person for passive
    @classmethod
    def addPassiveComplementsNumberPerson(cls, phrase, parent, realisedElement, verbElement):
        passiveNumber = None
        passivePerson = None
        currentElement = None
        splitVerb = None
        verbPhrase = phrase.getFeatureAsElement(InternalFeature.VERB_PHRASE)
        # count complements to set plural feature if more than one
        numComps = 0
        coordSubj = False
        if phrase.getFeatureAsBoolean(Feature.PASSIVE) and verbPhrase is not None and not \
                InterrogativeType.WHAT_OBJECT==phrase.getFeature(Feature.INTERROGATIVE_TYPE):
            # complements of a clause are stored in the VPPhraseSpec
            for subject in verbPhrase.getFeatureAsElementList(InternalFeature.COMPLEMENTS):
                if DiscourseFunction.OBJECT == subject.getFeature(InternalFeature.DISCOURSE_FUNCTION):
                    subject.setFeature(Feature.PASSIVE, True)
                    numComps += 1
                    currentElement = parent.realise(subject)
                    if currentElement is not None:
                        currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.OBJECT)
                        if phrase.hasFeature(Feature.INTERROGATIVE_TYPE):
                            splitVerb = currentElement
                        else:
                            realisedElement.addComponent(currentElement)
                    # flag if passive subject is coordinated with an "and"
                    if not coordSubj and isinstance(subject, CoordinatedPhraseElement):
                        conj = subject.getConjunction()
                        coordSubj = (conj is not None and conj=="and")
                    if passiveNumber is None:
                        passiveNumber = subject.getFeature(Feature.NUMBER)
                    else:
                        passiveNumber = NumberAgreement.PLURAL
                    if Person.FIRST == subject.getFeature(Feature.PERSON):
                        passivePerson = Person.FIRST
                    elif Person.SECOND == subject.getFeature(Feature.PERSON) and not \
                            Person.FIRST == passivePerson:
                        passivePerson = Person.SECOND
                    elif passivePerson is None:
                        passivePerson = Person.THIRD
                    if Form.GERUND == phrase.getFeature(Feature.FORM) and not \
                            phrase.getFeatureAsBoolean(Feature.SUPPRESS_GENITIVE_IN_GERUND):
                        subject.setFeature(Feature.POSSESSIVE, True)
        if verbElement is not None:
            if passivePerson is not None:
                verbElement.setFeature(Feature.PERSON, passivePerson)
            if numComps > 1 or coordSubj:
                verbElement.setFeature(Feature.NUMBER, NumberAgreement.PLURAL)
            elif passiveNumber is not None:
                verbElement.setFeature(Feature.NUMBER, passiveNumber)
        return splitVerb

    # Adds the subjects to the beginning of the clause unless the clause is
    # infinitive, imperative or passive, or the subjects split the verb.
    @classmethod
    def addSubjectsToFront(cls, phrase, parent, realisedElement, splitVerb):
        if not Form.INFINITIVE == phrase.getFeature(Feature.FORM) and not \
               Form.IMPERATIVE == phrase.getFeature(Feature.FORM) and not \
               phrase.getFeatureAsBoolean(Feature.PASSIVE) and splitVerb is None:
            realisedElement.addComponents(cls.realiseSubjects(phrase, parent).getChildren())

    # Realises the subjects for the clause.
    @classmethod
    def realiseSubjects(cls, phrase, parent):
        currentElement = None
        realisedElement = ListElement()
        for subject in phrase.getFeatureAsElementList(InternalFeature.SUBJECTS):
            subject.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.SUBJECT)
            if Form.GERUND==phrase.getFeature(Feature.FORM) and not \
                    phrase.getFeatureAsBoolean(Feature.SUPPRESS_GENITIVE_IN_GERUND):
                subject.setFeature(Feature.POSSESSIVE, True)
            currentElement = parent.realise(subject)
            if currentElement is not None:
                realisedElement.addComponent(currentElement)
        return realisedElement

    # This is the main controlling method for handling interrogative clauses.
    @classmethod
    def realiseInterrogative(cls, phrase, parent, realisedElement, phraseFactory, verbElement):
        splitVerb = None
        if phrase.getParent() is not None:
            phrase.getParent().setFeature(InternalFeature.INTERROGATIVE, True)
        ptype = phrase.getFeature(Feature.INTERROGATIVE_TYPE)
        if isinstance(ptype, InterrogativeType):
            if ptype == InterrogativeType.YES_NO:
                splitVerb = cls.realiseYesNo(phrase, parent, verbElement, phraseFactory, realisedElement)
            elif ptype == InterrogativeType.WHO_SUBJECT or ptype == InterrogativeType.WHAT_SUBJECT:
                cls.realiseInterrogativeKeyWord(ptype.getString(), LexicalCategory.PRONOUN, parent, \
                        realisedElement, phraseFactory)
                phrase.removeFeature(InternalFeature.SUBJECTS)
            elif ptype == InterrogativeType.HOW_MANY:
                cls.realiseInterrogativeKeyWord("how", LexicalCategory.PRONOUN, parent, realisedElement, \
                                            phraseFactory)
                cls.realiseInterrogativeKeyWord("many", LexicalCategory.ADVERB, parent, realisedElement, \
                                            phraseFactory)
            elif ptype in [InterrogativeType.HOW, InterrogativeType.WHY, InterrogativeType.WHERE, \
                           InterrogativeType.WHO_OBJECT, InterrogativeType.WHO_INDIRECT_OBJECT, \
                           InterrogativeType.WHAT_OBJECT]:
                splitVerb = cls.realiseObjectWHInterrogative(ptype.getString(), phrase, parent, realisedElement, \
                                                         phraseFactory)
            elif ptype == InterrogativeType.HOW_PREDICATE:
                splitVerb = cls.realiseObjectWHInterrogative("how", phrase, parent, realisedElement, phraseFactory)
        return splitVerb

    # Check if a sentence has an auxiliary (needed to relise questions correctly)
    @classmethod
    def hasAuxiliary(cls, phrase):
        return phrase.hasFeature(Feature.MODAL) or phrase.getFeatureAsBoolean(Feature.PERFECT) or \
               phrase.getFeatureAsBoolean(Feature.PROGRESSIVE) or \
               Tense.FUTURE == phrase.getFeature(Feature.TENSE)

    # Controls the realisation of wh object questions.
    @classmethod
    def realiseObjectWHInterrogative(cls, keyword, phrase, parent, realisedElement, phraseFactory):
        splitVerb = None
        cls.realiseInterrogativeKeyWord(keyword, LexicalCategory.PRONOUN, parent, realisedElement, \
                                        phraseFactory)
        if not cls.hasAuxiliary(phrase) and not VerbPhraseHelper.isCopular(phrase):
            cls.addDoAuxiliary(phrase, parent, phraseFactory, realisedElement)
        elif not phrase.getFeatureAsBoolean(Feature.PASSIVE):
            splitVerb = cls.realiseSubjects(phrase, parent)
        return splitVerb

    # Adds a do verb to the realisation of this clause.
    @classmethod
    def addDoAuxiliary(cls, phrase, parent, phraseFactory, realisedElement):
        doPhrase = phraseFactory.createVerbPhrase("do")
        doPhrase.setFeature(Feature.TENSE, phrase.getFeature(Feature.TENSE))
        doPhrase.setFeature(Feature.PERSON, phrase.getFeature(Feature.PERSON))
        doPhrase.setFeature(Feature.NUMBER, phrase.getFeature(Feature.NUMBER))
        realisedElement.addComponent(parent.realise(doPhrase))

    # Realises the key word of the interrogative. For example, who,
    # what
    @classmethod
    def realiseInterrogativeKeyWord(cls, keyWord, cat, parent, realisedElement, phraseFactory):
        if keyWord is not None:
            question = phraseFactory.createWord(keyWord, cat)
            currentElement = parent.realise(question)
            if currentElement is not None:
                realisedElement.addComponent(currentElement)

    # Performs the realisation for YES/NO types of questions.
    @classmethod
    def realiseYesNo(cls, phrase, parent, verbElement, phraseFactory, realisedElement):
        splitVerb = None
        if not (isinstance(verbElement, VPPhraseSpec) and VerbPhraseHelper.isCopular(verbElement.getVerb())) and not \
                phrase.getFeatureAsBoolean(Feature.PROGRESSIVE) and not phrase.hasFeature(Feature.MODAL) and not \
                Tense.FUTURE==phrase.getFeature(Feature.TENSE) and not \
                phrase.getFeatureAsBoolean(Feature.NEGATED) and not \
                phrase.getFeatureAsBoolean(Feature.PASSIVE):
            cls.addDoAuxiliary(phrase, parent, phraseFactory, realisedElement)
        else:
            splitVerb = cls.realiseSubjects(phrase, parent)
        return splitVerb

    # Realises the cue phrase for the clause if it exists.
    @classmethod
    def addCuePhrase(cls, phrase, parent, realisedElement):
        currentElement = parent.realise(phrase.getFeatureAsElement(Feature.CUE_PHRASE))
        if currentElement is not None:
            currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.CUE_PHRASE)
            realisedElement.addComponent(currentElement)

    # Checks to see if this clause is a subordinate clause.
    @classmethod
    def addComplementiser(cls, phrase, parent, realisedElement):
        if ClauseStatus.SUBORDINATE== phrase.getFeature(InternalFeature.CLAUSE_STATUS) and not \
                phrase.getFeatureAsBoolean(Feature.SUPRESSED_COMPLEMENTISER):
            currentElement = parent.realise(phrase.getFeatureAsElement(Feature.COMPLEMENTISER))
            if currentElement is not None:
                realisedElement.addComponent(currentElement)

    # Copies the front modifiers of the clause to the list of post-modifiers of
    # the verb only if the phrase has infinitive form.
    @classmethod
    def copyFrontModifiers(cls, phrase, verbElement):
        frontModifiers = phrase.getFeatureAsElementList(InternalFeature.FRONT_MODIFIERS)
        clauseForm = phrase.getFeature(Feature.FORM)
        # do not overwrite existing post-mods in the VP
        if verbElement is not None:
            phrasePostModifiers = phrase.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)
            if isinstance(verbElement, PhraseElement):
                verbPostModifiers = verbElement.getFeatureAsElementList(InternalFeature.POSTMODIFIERS)
                for eachModifier in phrasePostModifiers:
                    # need to check that VP doesn't already contain the post-modifier
                    # this only happens if the phrase has already been realised
                    # and later modified, with realiser called again. In that
                    # case, postmods will be copied over twice
                    if eachModifier not in verbPostModifiers:
                        verbElement.addPostModifier(eachModifier)
        if Form.INFINITIVE == clauseForm:
            phrase.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)
            for eachModifier in frontModifiers:
                if isinstance(verbElement, PhraseElement):
                    verbElement.addPostModifier(eachModifier)
            phrase.removeFeature(InternalFeature.FRONT_MODIFIERS)
            if verbElement is not None:
                verbElement.setFeature(InternalFeature.NON_MORPH, True)

    # Checks the discourse function of the clause and alters the form of the
    # clause as necessary.
    @classmethod
    def checkDiscourseFunction(cls, phrase):
        subjects = phrase.getFeatureAsElementList(InternalFeature.SUBJECTS)
        clauseForm = phrase.getFeature(Feature.FORM)
        discourseValue = phrase.getFeature(InternalFeature.DISCOURSE_FUNCTION)
        if discourseValue in [DiscourseFunction.OBJECT, DiscourseFunction.INDIRECT_OBJECT]:
            if Form.IMPERATIVE==clauseForm:
                phrase.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)
                phrase.setFeature(Feature.FORM, Form.INFINITIVE)
            elif Form.GERUND==clauseForm and len(subjects)==0:
                phrase.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)
        elif discourseValue==DiscourseFunction.SUBJECT:
            phrase.setFeature(Feature.FORM, Form.GERUND)
            phrase.setFeature(Feature.SUPRESSED_COMPLEMENTISER, True)

    # Checks the subjects of the phrase to determine if there is more than one
    # subject. This ensures that the verb phrase is correctly set. Also set
    # person correctly
    @classmethod
    def checkSubjectNumberPerson(cls, phrase, verbElement):
        currentElement = None
        subjects = phrase.getFeatureAsElementList(InternalFeature.SUBJECTS)
        pluralSubjects = False
        person = None
        if subjects:
            if len(subjects) == 1:
                currentElement = subjects[0]
                # coordinated NP with "and" are plural (not coordinated NP with "or")
                if isinstance(currentElement, CoordinatedPhraseElement) and currentElement.checkIfPlural():
                    pluralSubjects = True
                elif currentElement.getFeature(Feature.NUMBER)==NumberAgreement.PLURAL and not \
                        isinstance(currentElement, SPhraseSpec):
                    pluralSubjects = True
                elif currentElement.isA(PhraseCategory.NOUN_PHRASE):
                    currentHead = currentElement.getFeatureAsElement(InternalFeature.HEAD)
                    person = currentElement.getFeature(Feature.PERSON)
                    if currentHead is None:
                        # subject is null and therefore is not gonna be plural
                        pluralSubjects = False
                    elif currentHead.getFeature(Feature.NUMBER)==NumberAgreement.PLURAL:
                        pluralSubjects = True
                    elif isinstance(currentHead, ListElement):
                        pluralSubjects = True
            else:   # more than 1 subject
                pluralSubjects = True
        if verbElement is not None:
            plural = NumberAgreement.PLURAL if pluralSubjects else  phrase.getFeature(Feature.NUMBER)
            verbElement.setFeature(Feature.NUMBER,  plural)
            if person is not None:
                verbElement.setFeature(Feature.PERSON, person)



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the grammatical function that an element might
# take. class DiscourseFunction(Enum):
class DiscourseFunction(Enum):
    # Auxiliaries are the additional verbs added to a verb phrase to alter the
    # meaning being described.
    AUXILIARY = 0
    # Complements are additional components that are required to complement the
    # meaning of a sentence.
    COMPLEMENT = 1
    # A conjunction is a word that links items together in a coordinated
    # phrase. The most common conjunctions are and and but.
    CONJUNCTION = 2
    # Cue phrases are added to sentence to indicate document structure or flow.
    # They normally do not add any semantic information to the phrase.
    CUE_PHRASE = 3
    # Front modifiers are modifiers that apply to clauses. They are placed in
    # the syntactical structure after the cue phrase but before the subject.
    FRONT_MODIFIER = 4
    # This represents the main item of the phrase. For verb phrases, the head
    # will be the main verb. For noun phrases, the head will be the subject
    # noun. For adjective, adverb and prepositional phrases, the head will be
    # the adjective, adverb and preposition respectively.
    HEAD = 5
    # This is the indirect object of a verb phrase or an additional object that
    # is affected by the action performed. This is typically the recipient of
    # give.
    INDIRECT_OBJECT = 6
    # This is the object of a verb phrase and represents the item that the
    # action is performed upon.
    OBJECT = 7
    # Pre-modifiers, typically adjectives and adverbs, appear before the head
    # of a phrase. They can apply to noun phrases and verb phrases.
    PRE_MODIFIER = 8
     # Post-modifiers, typically adjectives and adverbs, are added after the
     # head of the phrase.
    POST_MODIFIER = 9
    # The specifier, otherwise known as the determiner, is a word that can be
    # placed before a noun in a noun phrase.
    SPECIFIER = 10
    # This is the subject of a verb phrase and represents the entity performing
    # the action.
    SUBJECT = 11
    # The verb phrase highlights the part of a clause that forms the verb
    # phrase. Verb phrases can be formed of a single verb or from a verb with a
    # particle, such as kiss, talk, bark,
    # fall down, pick up.
    VERB_PHRASE = 12

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum


# This is an enumeration of the two different types of clauses used in the
# SimplNLG package.
class ClauseStatus(Enum):
     # This enumeration represents a matrix clause. A matrix clause is not
     # subordinate to any other clause and therefore sits at the top-level of
     #the clause hierarchy, typically spanning the whole sentence.
    MATRIX = 0
     # The subordinate clauses are contained within a higher clause.
    SUBORDINATE = 1

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.


# This class defines a list of features internally used within the SimpleNLG
# system.
class InternalFeature(object):
    # This feature determines if the element is an acronym.
    ACRONYM = "acronym"
    # This feature is used to reference the base word element as created by the
    # lexicon.
    BASE_WORD = "base_word"
    # This feature determines the status of a sentence
    CLAUSE_STATUS = "clause_status"
    # This feature refers to the list of complements for the phrase.
    COMPLEMENTS = "complements"
    # This feature refers to the list of components in a list
    COMPONENTS = "components"
    # This feature is the list of coordinated phrases in a
    # CoordinatedPhraseElement.
    COORDINATES = "coordinates"
    # This feature defines the role each element plays in the structure of the
    # text. For example, the phrase John played football has
    # John as the subject, play as the base verb and
    # football as the complement.
    DISCOURSE_FUNCTION = "discourse_function"
    NON_MORPH = "non_morph"
    # This feature tracks any front modifiers in sentences. Front modifiers are
    # placed after the cue phrase but before the subject.
    FRONT_MODIFIERS = "front_modifiers"
    # This feature points to the head element in a phrase. The head element is
    # deemed to be the subject in a noun phrase, the verb in a verb phrase, the
    # adjective in an adjective phrase, the adverb in an adverb phrase or the
    # preposition in a preposition phrase. The PhraseElement has a
    # convenience method for getting and setting the head feature.
    HEAD = "head"
    # This flag is used to determine if the modal should be included in the
    # verb phrase.
    IGNORE_MODAL = "ignore_modal"
    # This flag determines if the sentence is interrogative or not.
    INTERROGATIVE = "interrogative"
    # This feature represents the list of post-modifier elements.
    # Post-modifiers are added to the end of phrases and coordinated phrases.
    POSTMODIFIERS = "postmodifiers"
    # This feature represents the list of premodifier elements. Premodifiers
    # are added to phrases before the head of the phrase, and to coordinated
    # phrases before the coordinates.
    PREMODIFIERS = "premodifiers"
    # This flag is used to define whether a noun phrase has had its specifier
    # raised. It is used in conjunction with the RAISE_SECIFIER
    # feature.
    RAISED = "raised"
    # This flag determines if auxiliary verbs should be realised in coordinated
    # verb phrases.
    REALISE_AUXILIARY = "realise_auxiliary"
    # This feature contains the specifier for a noun phrase. For example
    # the and my.
    SPECIFIER = "specifier"
    # This feature represents the list of subjects in a clause.
    SUBJECTS = "subjects"
    # This feature represents the verb phrase in a clause.
    VERB_PHRASE = "verb_phrase"



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement        import *
from .PhraseCategory    import *
from ..features.Feature import *


# This class defines an element for representing canned text within the
# SimpleNLG library. Once assigned a value, the string element should not be
# changed by any other processors.
class StringElement(NLGElement):
    def __init__(self, value):
        super().__init__()
        self.setCategory(PhraseCategory.CANNED_TEXT)
        self.setFeature(Feature.ELIDED, False)
        self.setRealisation(value)

    # The string element contains no children so this method will always return
    # an empty list.
    # @Override
    def getChildren(self):
        return []

    #@Override
    def __str__(self):
        return self.getRealisation()

    # @see simplenlg.framework.NLGElement#equals(java.lang.Object)
    # @Override
    def __eq__(self, obj):
        return super()==obj and isinstance(obj, StringElement) and self.realisationsMatch(obj)
    def _ne__(self, obj):
        return not self.__eq__(obj)
    def __hash__(self):
        return hash( super() )

    def realisationsMatch(self, obj):
        if  self.getRealisation() == None:
            return obj.getRealisation() == None
        else:
            return self.getRealisation() == obj.getRealisation()

    # @Override
    def printTree(self, indent=None):
        pstr = "StringElement: content=\"" + self.getRealisation() + '\"'
        features = self.getAllFeatures()
        if features is not None:
            pstr += ", features=" + str(features)
        pstr += '\n'
        return pstr



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.DiscourseFunction          import *
from ...features.Feature                    import *
from ...features.InternalFeature            import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.DocumentElement           import *
from ...framework.InflectedWordElement      import *
from ...framework.LexicalCategory           import *
from ...framework.ListElement               import *
from ...framework.NLGElement                import *
from ...framework.NLGModule                 import *
from ...framework.StringElement             import *
from ...framework.WordElement               import *
from .MorphologyRules                       import *


# This is the processor for handling morphology within the SimpleNLG.
class MorphologyProcessor(NLGModule):
    def __init__(self):
        super().__init__()

    # @Override
    def initialise(self):
        pass # Do nothing

    def realise(self, element):
        if isinstance(element, NLGElement):
            return self._realiseElement(element)
        elif isinstance(element, list):
            return self._realiseList(element)
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    # @Override
    def _realiseElement(self, element):
        realisedElement = None
        if isinstance(element, InflectedWordElement):
            realisedElement = self.doMorphology(element)
        elif isinstance(element, StringElement):
            realisedElement = element
        elif isinstance(element, WordElement):
            # AG: now retrieves the default spelling variant, not the baseform
            # String baseForm = ((WordElement) element).getBaseForm();
            defaultSpell = element.getDefaultSpellingVariant()
            if defaultSpell is not None:
                realisedElement = StringElement(defaultSpell)
        elif isinstance(element, DocumentElement):
            children = element.getChildren()
            element.setComponents(self.realise(children))
            realisedElement = element
        elif isinstance(element, ListElement):
            realisedElement = ListElement()
            realisedElement.addComponents(self.realise(element.getChildren()))
        elif isinstance(element, CoordinatedPhraseElement):
            children = element.getChildren()
            element.clearCoordinates()
            if children:
                element.addCoordinate(self.realise(children[0]))
                for i, child in enumerate(children):
                    if 0==i: continue
                    element.addCoordinate(self.realise(child))
                realisedElement = element
        elif element is None:
            realisedElement = element
        return realisedElement

    # This is the main method for performing the morphology.
    def doMorphology(self, element):
        realisedElement = None
        if element.getFeatureAsBoolean(InternalFeature.NON_MORPH):
            realisedElement = StringElement(element.getBaseForm())
            realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                    element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        else:
            baseWord = element.getFeatureAsElement(InternalFeature.BASE_WORD)
            if baseWord is None and self.lexicon is not None:
                baseWord = self.lexicon.lookupWord(element.getBaseForm())
            category = element.getCategory()
            if isinstance(category, LexicalCategory):
                if category == LexicalCategory.PRONOUN:
                    realisedElement = MorphologyRules.doPronounMorphology(element)
                elif category == LexicalCategory.NOUN:
                    realisedElement = MorphologyRules.doNounMorphology(element, baseWord)
                elif category == LexicalCategory.VERB:
                    realisedElement = MorphologyRules.doVerbMorphology(element, baseWord)
                elif category == LexicalCategory.ADJECTIVE:
                    realisedElement = MorphologyRules.doAdjectiveMorphology(element, baseWord)
                elif category == LexicalCategory.ADVERB:
                    realisedElement = MorphologyRules.doAdverbMorphology(element, baseWord)
                else:
                    realisedElement = StringElement(element.getBaseForm())
                    realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                        element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    # @Override
    def _realiseList(self, elements):
        realisedElements = []
        currentElement = None
        determiner = None
        prevElement = None
        if elements is not None:
            for eachElement in elements:
                currentElement = self._realiseElement(eachElement)
                if currentElement is not None:
                    #pass the discourse function and appositive features -- important for orth processor
                    currentElement.setFeature(Feature.APPOSITIVE, eachElement.getFeature(Feature.APPOSITIVE))
                    function = eachElement.getFeature(InternalFeature.DISCOURSE_FUNCTION)
                    if function is not None:
                        currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, function)
                    if prevElement is not None and isinstance(prevElement, StringElement) and \
                            isinstance(eachElement, InflectedWordElement) and \
                            eachElement.getCategory() == LexicalCategory.NOUN:
                        prevString = prevElement.getRealisation()
                        prevElement.setRealisation(DeterminerAgrHelper.checkEndsWithIndefiniteArticle(\
                                prevString,currentElement.getRealisation()))
                    realisedElements.append(currentElement)
                    if determiner is None and DiscourseFunction.SPECIFIER == \
                            currentElement.getFeature(InternalFeature.DISCOURSE_FUNCTION):
                        determiner = currentElement
                        determiner.setFeature(Feature.NUMBER, eachElement.getFeature(Feature.NUMBER))
                    elif determiner is not None:
                        if isinstance(currentElement, ListElement):
                            # list elements: ensure det matches first element
                            children = currentElement.getChildren()
                            firstChild = None if not children else children[0]
                            if firstChild is not None:
                                #AG: need to check if child is a coordinate
                                if isinstance(firstChild, CoordinatedPhraseElement):
                                    MorphologyRules.doDeterminerMorphology(determiner, \
                                        firstChild.getChildren()[0].getRealisation())
                                else:
                                    MorphologyRules.doDeterminerMorphology(determiner, firstChild.getRealisation())
                        else:
                            # everything else: ensure det matches realisation
                            MorphologyRules.doDeterminerMorphology(determiner, currentElement.getRealisation())
                        determiner = None
                prevElement = eachElement
        return realisedElements



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.Feature                    import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.DocumentElement           import *
from ...framework.InflectedWordElement      import *
from ...framework.LexicalCategory           import *
from ...framework.ListElement               import *
from ...framework.NLGElement                import *
from ...framework.NLGModule                 import *
from ...framework.PhraseCategory            import *
from ...framework.PhraseElement             import *
from ...framework.WordElement               import *
from .ClauseHelper                          import *
from .CoordinatedPhraseHelper               import *
from .NounPhraseHelper                      import *
from .PhraseHelper                          import *
from .VerbPhraseHelper                      import *

# This is the processor for handling syntax within the SimpleNLG. The processor
# translates phrases into lists of words.
class SyntaxProcessor(NLGModule):
    def __init__(self):
        super().__init__()

    # @Override
    def initialise(self):
        pass

    # @Override
    def realise(self, element):
        if element is None or element.getFeatureAsBoolean(Feature.ELIDED):
            return None
        elif isinstance(element, NLGElement):
            return self._realiseElement(element)
        elif isinstance(element, list):
            return self._realiseElementList(element)
        else:
            raise ValueError('Invalid element type: ' + str(type(element)))

    def _realiseElement(self, element):
        realisedElement = None
        if element is not None and not element.getFeatureAsBoolean(Feature.ELIDED):
            if isinstance(element, DocumentElement):
                children = element.getChildren()
                element.setComponents(self._realiseElementList(children))
                realisedElement = element
            elif isinstance(element, PhraseElement):
                realisedElement = self.realisePhraseElement(element)
            elif isinstance(element, ListElement):
                realisedElement = ListElement()
                realisedElement.addComponents(self._realiseElementList(element.getChildren()))
            elif isinstance(element, InflectedWordElement):
                baseForm = element.getBaseForm()
                category = element.getCategory()
                if self.lexicon is not None and baseForm is not None:
                    word = element.getBaseWord()
                    if word is None:
                        if isinstance(category, LexicalCategory):
                            word = self.lexicon.lookupWord(baseForm, category)
                        else:
                            word = self.lexicon.lookupWord(baseForm)
                    if word is not None:
                        element.setBaseWord(word)
                realisedElement = element
            elif isinstance(element, WordElement):
                # AG: need to check if it's a word element, in which case it
                # needs to be marked for inflection
                infl = InflectedWordElement(element)
                # # the inflected word inherits all features from the base word
                for feature in element.getAllFeatureNames():
                    infl.setFeature(feature, element.getFeature(feature))
                realisedElement = self._realiseElement(infl)
            elif isinstance(element, CoordinatedPhraseElement):
                realisedElement = CoordinatedPhraseHelper.realise(self, element)
            else:
                realisedElement = element
        # Remove the spurious ListElements that have only one element.
        if isinstance(realisedElement, ListElement):
            if len(realisedElement) == 1:
                realisedElement = realisedElement.getFirst()
        return realisedElement

    # @Override
    def _realiseElementList(self, elements):
        realisedList = []
        childRealisation = None
        if elements is not None:
            for eachElement in elements:
                if eachElement is not None:
                    childRealisation = self._realiseElement(eachElement)
                    if childRealisation is not None:
                        if isinstance(childRealisation, ListElement):
                            realisedList.extend(childRealisation.getChildren())
                        else:
                            realisedList.append(childRealisation)
        return realisedList

    # Realises a phrase element.
    def realisePhraseElement(self, phrase):
        realisedElement = None
        if phrase is not None:
            category = phrase.getCategory()
            if isinstance(category, PhraseCategory):
                if category == PhraseCategory.CLAUSE:
                    realisedElement = ClauseHelper.realise(self, phrase)
                elif category == PhraseCategory.NOUN_PHRASE:
                    realisedElement = NounPhraseHelper.realise(self, phrase)
                elif category == PhraseCategory.VERB_PHRASE:
                    realisedElement = VerbPhraseHelper.realise(self, phrase)
                elif category in [PhraseCategory.PREPOSITIONAL_PHRASE, PhraseCategory.ADJECTIVE_PHRASE, \
                        PhraseCategory.ADVERB_PHRASE]:
                    realisedElement = PhraseHelper.realise(self, phrase)
                else:
                    realisedElement = phrase
        return realisedElement



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

import re
from ...features.DiscourseFunction      import *
from ...features.Feature                import *
from ...features.Form                   import *
from ...features.Gender                 import *
from ...features.InternalFeature        import *
from ...features.LexicalFeature         import *
from ...features.NumberAgreement        import *
from ...features.Inflection             import *
from ...features.Person                 import *
from ...features.Tense                  import *
from ...framework.LexicalCategory       import *
from ...framework.StringElement         import *
from .DeterminerAgrHelper               import *


# This contains a number of rules for doing simple inflection.
class MorphologyRules(object):
    # A triple array of Pronouns organised by singular/plural,
    # possessive/reflexive/subjective/objective and by gender/person.
    PRONOUNS = [ [ [ "I", "you", "he", "she", "it" ],
                   [ "me", "you", "him", "her", "it" ],
                   [ "myself", "yourself", "himself", "herself", "itself" ],
                   [ "mine", "yours", "his", "hers", "its" ],
                   [ "my", "your", "his", "her", "its" ] ],
                 [ [ "we", "you", "they", "they", "they" ],
                   [ "us", "you", "them", "them", "them" ],
                   [ "ourselves", "yourselves", "themselves", "themselves", "themselves" ],
                   [ "ours", "yours", "theirs", "theirs", "theirs" ],
                   [ "our", "your", "their", "their", "their" ] ] ]
    WH_PRONOUNS = [ "who", "what", "which", "where", "why", "how", "how many" ]

    # This method performs the morphology for nouns.
    @classmethod
    def doNounMorphology(cls, element, baseWord):
        realised = ''
        # base form from baseWord if it exists, otherwise from element
        baseForm = cls.getBaseForm(element, baseWord)
        if element.isPlural() and not element.getFeatureAsBoolean(LexicalFeature.PROPER):
            pluralForm = None
            elementDefaultInfl = element.getFeature(LexicalFeature.DEFAULT_INFL)
            if elementDefaultInfl is not None and Inflection.UNCOUNT==elementDefaultInfl:
                pluralForm = baseForm
            else:
                pluralForm = element.getFeatureAsString(LexicalFeature.PLURAL)
            if pluralForm is None and baseWord is not None:
                baseDefaultInfl = baseWord.getFeatureAsString(LexicalFeature.DEFAULT_INFL)
                if baseDefaultInfl is not None and baseDefaultInfl == "uncount":
                    pluralForm = baseForm
                else:
                    pluralForm = baseWord.getFeatureAsString(LexicalFeature.PLURAL)
            if pluralForm is None:
                pattern = element.getFeature(LexicalFeature.DEFAULT_INFL)
                if Inflection.GRECO_LATIN_REGULAR == pattern:
                    pluralForm = cls.buildGrecoLatinPluralNoun(baseForm)
                else:
                    pluralForm = cls.buildRegularPluralNoun(baseForm)
            realised = pluralForm
        else:
            realised = baseForm
        realised = cls.checkPossessive(element, realised)
        realisedElement = StringElement(realised)
        realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    # Builds a plural for regular nouns. The rules are performed in this order:
    @classmethod
    def buildRegularPluralNoun(cls, baseForm):
        plural = None
        if baseForm is not None:
            if re.fullmatch(r".*[^aeiou]y\b", baseForm):
                plural = re.sub(r"y\b", "ies", baseForm)
            elif re.fullmatch(r".*([szx]|[cs]h)\b", baseForm):
                plural = baseForm + "es"
            else:
                plural = baseForm + "s"
        return plural

    # Builds a plural for Greco-Latin regular nouns.
    @classmethod
    def buildGrecoLatinPluralNoun(cls, baseForm):
        if not baseForm:
            return None
        if baseForm.endswith("us"):
            plural = re.sub(r"us\b", "i", baseForm)
        elif baseForm.endswith("ma"):
            plural = baseForm + "ta"
        elif baseForm.endswith("a"):
            plural = baseForm + "e"
        elif re.fullmatch(r".*(um|on)\b", baseForm):
            plural = re.sub(r"(um|on)\b",  "a", baseForm)
        elif baseForm.endswith("sis"):
            plural = re.sub(r"sis\b", "ses", baseForm)
        elif baseForm.endswith("is"):
            plural = re.sub(r"is\b", "ides", baseForm)
        elif baseForm.endswith("men"):
            plural = re.sub(r"men\b", "mina", baseForm)
        elif baseForm.endswith("ex"):
            plural = re.sub(r"ex\b", "ices", baseForm)
        elif baseForm.endswith("x"):
            plural = re.sub(r"x\b", "ces", baseForm)
        else:
            plural = baseForm
        return plural

    # This method performs the morphology for verbs.
    @classmethod
    def doVerbMorphology(cls, element, baseWord):
        realised = None
        numberValue = element.getFeature(Feature.NUMBER)
        personValue = element.getFeature(Feature.PERSON)
        tense       = element.getFeature(Feature.TENSE)
        if isinstance(tense, Tense):
            tenseValue = tense
        else:
            tenseValue = Tense.PRESENT
        formValue = element.getFeature(Feature.FORM)
        patternValue = element.getFeature(LexicalFeature.DEFAULT_INFL)
        # base form from baseWord if it exists, otherwise from element
        baseForm = cls.getBaseForm(element, baseWord)
        if element.getFeatureAsBoolean(Feature.NEGATED) or Form.BARE_INFINITIVE == formValue:
            realised = baseForm
        elif Form.PRESENT_PARTICIPLE == formValue:
            realised = element.getFeatureAsString(LexicalFeature.PRESENT_PARTICIPLE)
            if realised is None and baseWord is not None:
                realised = baseWord.getFeatureAsString(LexicalFeature.PRESENT_PARTICIPLE)
            if realised is None:
                if Inflection.REGULAR_DOUBLE == patternValue:
                    realised = cls.buildDoublePresPartVerb(baseForm)
                else:
                    realised = cls.buildRegularPresPartVerb(baseForm)
        elif Tense.PAST == tenseValue or Form.PAST_PARTICIPLE == formValue:
            if Form.PAST_PARTICIPLE == formValue:
                realised = element.getFeatureAsString(LexicalFeature.PAST_PARTICIPLE)
                if realised is None and baseWord is not None:
                    realised = baseWord.getFeatureAsString(LexicalFeature.PAST_PARTICIPLE)
                if realised is None:
                    if "be" == baseForm.lower():
                        realised = "been"
                    elif Inflection.REGULAR_DOUBLE == patternValue:
                        realised = cls.buildDoublePastVerb(baseForm)
                    else:
                        realised = cls.buildRegularPastVerb(baseForm, numberValue, personValue)
            else:
                realised = element.getFeatureAsString(LexicalFeature.PAST)
                if realised is None and baseWord is not None:
                    realised = baseWord.getFeatureAsString(LexicalFeature.PAST)
                if realised is None:
                    if Inflection.REGULAR_DOUBLE == patternValue:
                        realised = cls.buildDoublePastVerb(baseForm)
                    else:
                        realised = cls.buildRegularPastVerb(baseForm, numberValue, personValue)
        elif (numberValue is None or NumberAgreement.SINGULAR==numberValue) and \
                (personValue is None or Person.THIRD==personValue) and \
                (tenseValue is None or Tense.PRESENT==tenseValue):
            realised = element.getFeatureAsString(LexicalFeature.PRESENT3S)
            if realised is None and baseWord is not None and not "be" == baseForm.lower():
                realised = baseWord.getFeatureAsString(LexicalFeature.PRESENT3S)
            if realised is None:
                realised = cls.buildPresent3SVerb(baseForm)
        else:
            if "be"==baseForm.lower():
                if Person.FIRST == personValue and (NumberAgreement.SINGULAR==numberValue or numberValue is None):
                    realised = "am"
                else:
                    realised = "are"
            else:
                realised = baseForm
        realisedElement = StringElement(realised)
        realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    # return the base form of a word
    @classmethod
    def getBaseForm(cls, element, baseWord):
        # unclear what the right behaviour should be
        # for now, prefer baseWord.getBaseForm() to element.getBaseForm() for
        # verbs (ie, "is" mapped to "be")
        # but prefer element.getBaseForm() to baseWord.getBaseForm() for other
        # words (ie, "children" not mapped to "child")
        # AG: changed this to get the default spelling variant
        # needed to preserve spelling changes in the VP
        if LexicalCategory.VERB == element.getCategory():
            if baseWord is not None and baseWord.getDefaultSpellingVariant() is not None:
                return baseWord.getDefaultSpellingVariant()
            else:
                return element.getBaseForm()
        else:
            if element.getBaseForm() is not None:
                return element.getBaseForm()
            elif baseWord is None:
                return None
            else:
                return baseWord.getDefaultSpellingVariant()

    # Checks to see if the noun is possessive. If it is then nouns in ending in
    # -s become -s' while every other noun has -'s appended to
    # the end.
    @classmethod
    def checkPossessive(cls, element, realised):
        if element.getFeatureAsBoolean(Feature.POSSESSIVE):
            if realised[-1] == 's':
                realised += '\''
            else:
                realised += "'s"
        return realised

    # Builds the third-person singular form for regular verbs. The rules are
    # performed in this order:
    @classmethod
    def buildPresent3SVerb(cls, baseForm):
        morphology = None
        if baseForm is not None:
            if baseForm.lower() == "be":
                morphology = "is"
            elif re.fullmatch(r".*[szx(ch)(sh)]\b", baseForm):
                morphology = baseForm + "es"
            elif re.fullmatch(r".*[^aeiou]y\b", baseForm):
                morphology = re.sub(r"y\b", "ies", baseForm)
            else:
                morphology = baseForm + "s"
        return morphology

    # Builds the past-tense form for regular verbs. The rules are performed in
    # this order:
    @classmethod
    def buildRegularPastVerb(cls, baseForm, number, person):
        morphology = None
        if baseForm is not None:
            if baseForm.lower() == "be":
                if NumberAgreement.PLURAL == number:
                    morphology = "were"
                elif Person.SECOND == person:
                    morphology = "were"
                else:
                    morphology = "was"
            elif baseForm.endswith("e"):
                morphology = baseForm + "d"
            elif re.fullmatch(r".*[^aeiou]y\b",  baseForm):
                morphology = re.sub(r"y\b", "ied", baseForm)
            else:
                morphology = baseForm + "ed"
        return morphology

    # Builds the past-tense form for verbs that follow the doubling form of the
    # last consonant. -ed is added to the end after the last consonant
    # is doubled. For example, tug becomes tugged.
    @classmethod
    def buildDoublePastVerb(cls, baseForm):
        morphology = None
        if baseForm is not None:
            morphology = baseForm + baseForm[-1] + "ed"
        return morphology

    # Builds the present participle form for regular verbs. The rules are
    # performed in this order:
    @classmethod
    def buildRegularPresPartVerb(cls, baseForm):
        morphology = None
        if baseForm is not None:
            if baseForm.lower() == "be":
                morphology = "being"
            elif baseForm.endswith("ie"):
                morphology = re.sub(r"ie\b", "ying", baseForm)
            elif re.fullmatch(r".*[^iyeo]e\b", baseForm):
                morphology = re.sub(r"e\b", "ing", baseForm)
            else:
                morphology = baseForm + "ing"
        return morphology

    # Builds the present participle form for verbs that follow the doubling
    # form of the last consonant. -ing is added to the end after the
    # last consonant is doubled. For example, tug becomes
    # tugging.
    @classmethod
    def buildDoublePresPartVerb(cls, baseForm):
        morphology = None
        if baseForm is not None:
            morphology = baseForm + baseForm[-1] + "ing"
        return morphology

    # This method performs the morphology for adjectives.
    @classmethod
    def doAdjectiveMorphology(cls, element, baseWord):
        realised = None
        patternValue = element.getFeature(LexicalFeature.DEFAULT_INFL)
        # base form from baseWord if it exists, otherwise from element
        baseForm = cls.getBaseForm(element, baseWord)
        if element.getFeatureAsBoolean(Feature.IS_COMPARATIVE):
            realised = element.getFeatureAsString(LexicalFeature.COMPARATIVE)
            if realised is None and baseWord is not None:
                realised = baseWord.getFeatureAsString(LexicalFeature.COMPARATIVE)
            if realised is None:
                if Inflection.REGULAR_DOUBLE==patternValue:
                    realised = cls.buildDoubleCompAdjective(baseForm)
                else:
                    realised = cls.buildRegularComparative(baseForm)
        elif element.getFeatureAsBoolean(Feature.IS_SUPERLATIVE):
            realised = element.getFeatureAsString(LexicalFeature.SUPERLATIVE)
            if realised is None and baseWord is not None:
                realised = baseWord.getFeatureAsString(LexicalFeature.SUPERLATIVE)
            if realised is None:
                if Inflection.REGULAR_DOUBLE == patternValue:
                    realised = cls.buildDoubleSuperAdjective(baseForm)
                else:
                    realised = cls.buildRegularSuperlative(baseForm)
        else:
            realised = baseForm
        realisedElement = StringElement(realised)
        realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    # Builds the comparative form for adjectives that follow the doubling form
    # of the last consonant. -er is added to the end after the last
    # consonant is doubled. For example, fat becomes fatter.
    @classmethod
    def buildDoubleCompAdjective(cls, baseForm):
        morphology = None
        if baseForm is not None:
            morphology = baseForm + baseForm[-1] + "er"
        return morphology

    # Builds the comparative form for regular adjectives. The rules are
    # performed in this order:
    @classmethod
    def buildRegularComparative(cls, baseForm):
        morphology = None
        if baseForm is not None:
            if re.fullmatch(r".*[^aeiou]y\b", baseForm):
                morphology = re.sub(r"y\b", "ier", baseForm)
            elif baseForm.endswith("e"):
                morphology = baseForm + "r"
            else:
                morphology = baseForm + "er"
        return morphology

    # Builds the superlative form for adjectives that follow the doubling form
    # of the last consonant.
    @classmethod
    def buildDoubleSuperAdjective(cls, baseForm):
        morphology = None
        if baseForm is not None:
            morphology = baseForm + baseForm[-1] + "est"
        return morphology

    # Builds the superlative form for regular adjectives. The rules are
    # performed in this order:
    @classmethod
    def buildRegularSuperlative(cls, baseForm):
        morphology = None
        if baseForm is not None:
            if re.fullmatch(r".*[^aeiou]y\b", baseForm):
                morphology = re.sub(r"y\b", "iest", baseForm)
            elif baseForm.endswith("e"):
                morphology = baseForm + "st"
            else:
                morphology = baseForm + "est"
        return morphology

    # This method performs the morphology for adverbs.
    @classmethod
    def doAdverbMorphology(cls, element, baseWord):
        realised = None
        # base form from baseWord if it exists, otherwise from element
        baseForm = cls.getBaseForm(element, baseWord)
        if element.getFeatureAsBoolean(Feature.IS_COMPARATIVE):
            realised = element.getFeatureAsString(LexicalFeature.COMPARATIVE)
            if realised is None and baseWord is not None:
                realised = baseWord.getFeatureAsString(LexicalFeature.COMPARATIVE)
            if realised is None:
                realised = cls.buildRegularComparative(baseForm)
        elif element.getFeatureAsBoolean(Feature.IS_SUPERLATIVE):
            realised = element.getFeatureAsString(LexicalFeature.SUPERLATIVE)
            if realised is None and baseWord is not None:
                realised = baseWord.getFeatureAsString(LexicalFeature.SUPERLATIVE)
            if realised is None:
                realised = cls.buildRegularSuperlative(baseForm)
        else:
            realised = baseForm
        realisedElement = StringElement(realised)
        realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, \
                element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    # This method performs the morphology for pronouns.
    @classmethod
    def doPronounMorphology(cls, element):
        realised = None
        if not element.getFeatureAsBoolean(InternalFeature.NON_MORPH) and not cls.isWHPronoun(element):
            genderValue = element.getFeature(LexicalFeature.GENDER)
            personValue = element.getFeature(Feature.PERSON)
            discourseValue = element.getFeature(InternalFeature.DISCOURSE_FUNCTION)
            if element.isPlural():
                numberIndex = 1
            else:
                numberIndex = 0
            if isinstance(genderValue, Gender):
                genderIndex = genderValue.value
            else:
                genderIndex = 2
            if isinstance(personValue, Person):
                personIndex = personValue.value
            else:
                personIndex = 2
            if personIndex == 2:
                personIndex += genderIndex
            positionIndex = 0
            if element.getFeatureAsBoolean(LexicalFeature.REFLEXIVE):
                positionIndex = 2
            elif element.getFeatureAsBoolean(Feature.POSSESSIVE):
                positionIndex = 3
                if DiscourseFunction.SPECIFIER == discourseValue:
                    positionIndex += 1
            else:
                if (DiscourseFunction.SUBJECT==discourseValue and not element.getFeatureAsBoolean(Feature.PASSIVE)) or \
                   (DiscourseFunction.OBJECT ==discourseValue and     element.getFeatureAsBoolean(Feature.PASSIVE)) or \
                   DiscourseFunction.SPECIFIER == discourseValue or \
                   (DiscourseFunction.COMPLEMENT == discourseValue and element.getFeatureAsBoolean(Feature.PASSIVE)):
                    positionIndex = 0
                else:
                    positionIndex = 1
            realised = cls.PRONOUNS[numberIndex][positionIndex][personIndex]
        else:
            realised = element.getBaseForm()
        realisedElement = StringElement(realised)
        realisedElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, element.getFeature(InternalFeature.DISCOURSE_FUNCTION))
        return realisedElement

    @classmethod
    def isWHPronoun(cls, word):
        base = word.getBaseForm()
        wh = False
        if base is not None:
            for pronoun in cls.WH_PRONOUNS:
                wh = pronoun == base
                if wh:
                    break
        return wh

    # This method performs the morphology for determiners.
    @classmethod
    def doDeterminerMorphology(cls, determiner, realisation):
        if realisation is not None:
            if determiner.getRealisation() == "a":
                if determiner.isPlural():
                    determiner.setRealisation("some")
                elif DeterminerAgrHelper.requiresAn(realisation):
                    determiner.setRealisation("an")



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.


# This class keeps track of the prefix for numbered lists.
class NumberedPrefix(object):
    def __init__(self):
        self.prefix = "0"

    def increment(self):
        dotPosition = self.prefix.rfind('.')
        if dotPosition < 0:
            counter = int(self.prefix)
            counter += 1
            self.prefix = str(counter)
        else:
            subCounterStr = self.prefix[dotPosition+1:]
            subCounter = int(subCounterStr)
            subCounter += 1
            self.prefix = self.prefix[:dotPosition] + "." + str(subCounter)

    # This method starts a new level to the prefix
    def upALevel(self):
        if self.prefix == "0":
            self.prefix = "1"
        else:
            self.prefix += ".1"

    # This method removes a level from the prefix .
    def downALevel(self):
        dotPosition = self.prefix.rfind('.')
        if dotPosition < 0:
            self.prefix = "0"
        else:
            self.prefix = self.prefix[:dotPosition]

    def getPrefix(self):
        return self.prefix

    def setPrefix(self, prefix):
        self.prefix = prefix



import os
resource_directory = os.path.join(os.path.dirname(__file__), 'resources')

from .features      import *
from .format        import *
from .framework     import *
from .lexicon       import *
from .morphology    import *
from .orthography   import *
from .phrasespec    import *
from .realiser      import *
from .syntax        import *



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

# This class defines a list of features which can be set up users of SimpleNLG.
# Note that there are three feature classes in SimpleNLG.
class Feature(object):
    # This feature determines if the adjectives should be reordered. Some
    # lexicons might support the positioning of adjectives and by default, this
    # order will be used. The user can override the ordering of adjectives by
    # setting this
    ADJECTIVE_ORDERING = "adjective_ordering"
    # This feature determines if the auxiliary verbs in a clause should be
    # aggregated.
    AGGREGATE_AUXILIARY = "aggregate_auxiliary"
    # The complementiser is the word that joins a subordinate clause to the
    # parent clause.
    COMPLEMENTISER = "complementiser"
    # This feature represents the word (or phrase) used for linking coordinated
    CONJUNCTION = "conjunction"
    # This feature represents the type of conjunction this coordination
    # represents.
    CONJUNCTION_TYPE = "conjunction_type"
    # An appositive is a type of postmodifying phrase which is quasi-synonymous
    # with, or a possible replacement of, the phrase it modifies. A typical
    # example occurs with NPs
    APPOSITIVE = "appositive"
    # This feature represents the cue phrase of a sentence. Cue phrases
    # sometimes appear at the start of sentences.
    CUE_PHRASE = "cue_phrase"
    # This features determines if the phrase is elided. Elided phrases are
    # omitted from the final realisation.
    ELIDED = "elided"
    # This feature dictates the form that a verb takes.
    FORM = "form"
    # This feature determines the type of interrogative (question) used for the
    # clause.
    INTERROGATIVE_TYPE = "interrogative_type"
    # This flag determines if the Adjective or Adverb should be inflected into
    # the comparative form.
    IS_COMPARATIVE = "is_comparative"
    # This flag determines if the Adjective or Adverb should be inflected into
    # the superlative form.
    IS_SUPERLATIVE = "is_superlative"
    # The modal represents the modal auxiliary verb that is used in a verb
    # phrase to express mood or tense.
    MODAL = "modal"
    # The flag determines if the corresponding verb phrase should be negated.
    NEGATED = "negated"
    # This feature is used to determine if the element is to be represented in
    # singular or plural form.
    NUMBER = "number"
    # This feature represents the pattern that a particular word follows
    # when being inflected.
    PATTERN = "pattern"    # note -commented out in original code
    # This feature represents a particle used in conjunction with a verb.
    PARTICLE = "particle"
    # This flag shows if the phrase or clause should be written in the passive
    # form.
    PASSIVE = "passive"
    # This flag shows if the phrase or clause should be written in the perfect
    # form.
    PERFECT = "perfect"
    # This feature represents the first-person, second-person or third-person
    # nature of the phrase.
    PERSON = "person"
    # This flag shows if the noun phrase should be written in the possessive
    # form.
    POSSESSIVE = "possessive"
    # This flag can be set for noun phrases where it is desirable to overwrite
    # the subject with a suitable pronoun.
    PRONOMINAL = "pronominal"
    # This flag determines if the verb phrase should be constructed in the
    # progressive form.
    PROGRESSIVE = "progressive"
    # This flag can be set when specifiers in a coordinated phrase should be
    # raised.
    RAISE_SPECIFIER = "raise_specifier"
    # This flag is set when it is necessary to suppress the genitive when
    # dealing with gerund forms.
    SUPPRESS_GENITIVE_IN_GERUND = "suppress_genitive_in_gerund"
    # This flag determines if complementisers in subordinating clauses should
    # be suppressed.
    SUPRESSED_COMPLEMENTISER = "suppressed_complementiser"
    # This flag represents the tense or temporal quality of a verb.
    TENSE = "tense"



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# An enumeration representing the different types of temporal sense that can be
# applied to a verb.
class Tense(Enum):
    # The action described by the verb will happen in the future. For example,
    # John will kiss Mary, the dog will eat a bone.
    FUTURE = 0
    # The action described by the verb happened in the past. For example,
    # John kissed Mary, the dog ate a bone.
    PAST = 1
    # The action described by the verb is happening in the present time. For
    # example, John kisses Mary, the dog eats a bone.
    PRESENT = 2

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from .NLGElement                import *
from ..features.InternalFeature import *
from ..features.LexicalFeature  import *
from .WordElement               import *


# This class defines the NLGElement that is used to represent an
# word that requires inflection by the morphology.
class InflectedWordElement(NLGElement):
    def __init__(self, word, category=None):
        super().__init__()
        if isinstance(word, WordElement):
            self.setFeature(InternalFeature.BASE_WORD, word)
            # AG: changed to use the default spelling variant
            # setFeature(LexicalFeature.BASE_FORM, word.getBaseForm());
            defaultSpelling = word.getDefaultSpellingVariant()
            self.setFeature(LexicalFeature.BASE_FORM, defaultSpelling)
            self.setCategory(word.getCategory())
        elif isinstance(word, str) and isinstance(category, LexicalCategory):
            self.setFeature(LexicalFeature.BASE_FORM, word)
            self.setCategory(category)
        else:
            raise ValueError('Invalid param types: ' + str(type(word)) + ', ' + str(type(category)))

    # This method returns null as the inflected word has no child components.
    # @Override
    def getChildren(self):
        return None

    # @Override
    def __str__(self):
        return "InflectedWordElement[" + self.getBaseForm() + ':' + str(self.getCategory()) + ']'

    # @Override
    def printTree(self, indent=None):
        pstr = "InflectedWordElement: base=" + self.getBaseForm() + ", category=" + \
               str(self.getCategory()) + ", " + super().__str__() + '\n'
        return pstr

    # Retrieves the base form for this element. The base form is the originally
    # supplied word.
    def getBaseForm(self):
        return self.getFeatureAsString(LexicalFeature.BASE_FORM)

    # Sets the base word for this element.
    def setBaseWord(self, word):
        self.setFeature(InternalFeature.BASE_WORD, word)

    # Retrieves the base word for this element.
    def getBaseWord(self):
        baseWord = self.getFeatureAsElement(InternalFeature.BASE_WORD)
        if isinstance(baseWord, WordElement):
            return baseWord
        return None



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ...features.DiscourseFunction          import *
from ...features.Feature                    import *
from ...features.Form                       import *
from ...features.InternalFeature            import *
from ...features.InterrogativeType          import *
from ...features.NumberAgreement            import *
from ...features.Tense                      import *
from ...framework.CoordinatedPhraseElement  import *
from ...framework.InflectedWordElement      import *
from ...framework.LexicalCategory           import *
from ...framework.ListElement               import *
from ...framework.NLGElement                import *
from ...framework.PhraseCategory            import *
from ...framework.PhraseElement             import *
from ...framework.StringElement             import *
from ...framework.WordElement               import *
from ...phrasespec.SPhraseSpec              import *
from .PhraseHelper                          import *


# This class contains static methods to help the syntax processor realise verb
# phrases. It adds auxiliary verbs into the element tree as required.
class VerbPhraseHelper(object):
    # The main method for realising verb phrases.
    @classmethod
    def realise(cls, parent,  phrase):
        realisedElement = ListElement()
        mainVerbRealisation = []
        auxiliaryRealisation = []
        if phrase is not None:
            vgComponents = cls.createVerbGroup(parent, phrase)
            cls.splitVerbGroup(vgComponents, mainVerbRealisation, auxiliaryRealisation)
            realisedElement = ListElement()
            if not phrase.hasFeature(InternalFeature.REALISE_AUXILIARY) or \
                    phrase.getFeatureAsBoolean(InternalFeature.REALISE_AUXILIARY):
                cls.realiseAuxiliaries(parent, realisedElement, auxiliaryRealisation)
                PhraseHelper.realiseList(parent, realisedElement, phrase.getPreModifiers(), \
                    DiscourseFunction.PRE_MODIFIER)
                cls.realiseMainVerb(parent, phrase, mainVerbRealisation, realisedElement)
            elif cls.isCopular(phrase.getHead()):
                cls.realiseMainVerb(parent, phrase, mainVerbRealisation, realisedElement)
                PhraseHelper.realiseList(parent, realisedElement, phrase.getPreModifiers(), \
                        DiscourseFunction.PRE_MODIFIER)
            else:
                PhraseHelper.realiseList(parent, realisedElement, phrase.getPreModifiers(), \
                        DiscourseFunction.PRE_MODIFIER)
                cls.realiseMainVerb(parent, phrase, mainVerbRealisation, realisedElement)
            cls.realiseComplements(parent, phrase, realisedElement)
            PhraseHelper.realiseList(parent, realisedElement, phrase.getPostModifiers(), \
                    DiscourseFunction.POST_MODIFIER)
        return realisedElement

    # Realises the auxiliary verbs in the verb group.
    @classmethod
    def realiseAuxiliaries(cls, parent, realisedElement, auxiliaryRealisation):
        aux = None
        currentElement = None
        while auxiliaryRealisation:
            aux = auxiliaryRealisation.pop()
            currentElement = parent.realise(aux)
            if currentElement is not None:
                realisedElement.addComponent(currentElement)
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.AUXILIARY)

    # Realises the main group of verbs in the phrase.
    @classmethod
    def realiseMainVerb(cls, parent, phrase, mainVerbRealisation, realisedElement):
        currentElement = None
        main = None
        while mainVerbRealisation:
            main = mainVerbRealisation.pop()
            main.setFeature(Feature.INTERROGATIVE_TYPE, phrase.getFeature(Feature.INTERROGATIVE_TYPE))
            currentElement = parent.realise(main)
            if currentElement is not None:
                realisedElement.addComponent(currentElement)

    # Realises the complements of this phrase.
    @classmethod
    def realiseComplements(cls, parent, phrase, realisedElement):
        indirects = ListElement()
        directs = ListElement()
        unknowns = ListElement()
        discourseValue = None
        currentElement = None
        for complement in phrase.getFeatureAsElementList(InternalFeature.COMPLEMENTS):
            discourseValue = complement.getFeature(InternalFeature.DISCOURSE_FUNCTION)
            currentElement = parent.realise(complement)
            if currentElement is not None:
                currentElement.setFeature(InternalFeature.DISCOURSE_FUNCTION, DiscourseFunction.COMPLEMENT)
                if DiscourseFunction.INDIRECT_OBJECT==discourseValue:
                    indirects.addComponent(currentElement)
                elif DiscourseFunction.OBJECT == discourseValue:
                    directs.addComponent(currentElement)
                else:
                    unknowns.addComponent(currentElement)
        if not InterrogativeType.isIndirectObject(phrase.getFeature(Feature.INTERROGATIVE_TYPE)):
            realisedElement.addComponents(indirects.getChildren())
        if not phrase.getFeatureAsBoolean(Feature.PASSIVE):
            if not InterrogativeType.isObject(phrase.getFeature(Feature.INTERROGATIVE_TYPE)):
                realisedElement.addComponents(directs.getChildren())
            realisedElement.addComponents(unknowns.getChildren())

    # Splits the stack of verb components into two sections. One being the verb
    # associated with the main verb group, the other being associated with the
    # auxiliary verb group.
    @classmethod
    def splitVerbGroup(cls, vgComponents, mainVerbRealisation, auxiliaryRealisation):
        mainVerbSeen = False
        for word in vgComponents:
            if not mainVerbSeen:
                mainVerbRealisation.append(word)
                if word != "not":
                    mainVerbSeen = True
            else:
                auxiliaryRealisation.append(word)

    # Creates a stack of verbs for the verb phrase. Additional auxiliary verbs
    # are added as required based on the features of the verb phrase.
    @classmethod
    def createVerbGroup(cls, parent, phrase):
        actualModal = None
        formValue = phrase.getFeature(Feature.FORM)
        tenseValue = phrase.getFeature(Feature.TENSE)
        modal = phrase.getFeatureAsString(Feature.MODAL)
        modalPast = False
        vgComponents = []
        interrogative = phrase.hasFeature(Feature.INTERROGATIVE_TYPE)
        if Form.GERUND==formValue or Form.INFINITIVE==formValue:
            tenseValue = Tense.PRESENT
        if Form.INFINITIVE==formValue:
            actualModal = "to"
        elif formValue is None or Form.NORMAL==formValue:
            if Tense.FUTURE==tenseValue and modal is None and \
                    ((not isinstance(phrase.getHead(), CoordinatedPhraseElement)) or \
                    (isinstance(phrase.getHead(), CoordinatedPhraseElement) and interrogative)):
                actualModal = "will"
            elif modal is not None:
                actualModal = modal
                if Tense.PAST==tenseValue:
                    modalPast = True
        cls.pushParticles(phrase, parent, vgComponents)
        frontVG = cls.grabHeadVerb(phrase, tenseValue, modal is not None)
        cls.checkImperativeInfinitive(formValue, frontVG)
        if phrase.getFeatureAsBoolean(Feature.PASSIVE):
            frontVG = cls.addBe(frontVG, vgComponents, Form.PAST_PARTICIPLE)
        if phrase.getFeatureAsBoolean(Feature.PROGRESSIVE):
            frontVG = cls.addBe(frontVG, vgComponents, Form.PRESENT_PARTICIPLE)
        if phrase.getFeatureAsBoolean(Feature.PERFECT) or modalPast:
            frontVG = cls.addHave(frontVG, vgComponents, modal, tenseValue)
        frontVG = cls.pushIfModal(actualModal is not None, phrase, frontVG, vgComponents)
        frontVG = cls.createNot(phrase, vgComponents, frontVG, modal is not None)
        if frontVG is not None:
            cls.pushFrontVerb(phrase, vgComponents, frontVG, formValue, interrogative)
        cls.pushModal(actualModal, phrase, vgComponents)
        return vgComponents

    # Pushes the modal onto the stack of verb components.
    @classmethod
    def pushModal(cls, actualModal, phrase, vgComponents):
        if actualModal is not None and not phrase.getFeatureAsBoolean(InternalFeature.IGNORE_MODAL):
            vgComponents.append(InflectedWordElement(actualModal, LexicalCategory.MODAL))

    # Pushes the front verb onto the stack of verb components.
    @classmethod
    def pushFrontVerb(cls, phrase, vgComponents,  frontVG, formValue, interrogative):
        interrogType = phrase.getFeature(Feature.INTERROGATIVE_TYPE)
        if Form.GERUND == formValue:
            frontVG.setFeature(Feature.FORM, Form.PRESENT_PARTICIPLE)
            vgComponents.append(frontVG)
        elif Form.PAST_PARTICIPLE==formValue:
            frontVG.setFeature(Feature.FORM, Form.PAST_PARTICIPLE)
            vgComponents.append(frontVG)
        elif Form.PRESENT_PARTICIPLE==formValue:
            frontVG.setFeature(Feature.FORM, Form.PRESENT_PARTICIPLE)
            vgComponents.append(frontVG)
        elif (not (formValue is None or Form.NORMAL==formValue) or interrogative) and \
                not cls.isCopular(phrase.getHead()) and not vgComponents:
            # AG: fix below: if interrogative, only set non-morph feature in
            # case it's not WHO_SUBJECT OR WHAT_SUBJECT
            if not (InterrogativeType.WHO_SUBJECT==interrogType or InterrogativeType.WHAT_SUBJECT==interrogType):
                frontVG.setFeature(InternalFeature.NON_MORPH, True)
            vgComponents.append(frontVG)
        else:
            numToUse = cls.determineNumber(phrase.getParent(), phrase)
            frontVG.setFeature(Feature.TENSE, phrase.getFeature(Feature.TENSE))
            frontVG.setFeature(Feature.PERSON, phrase.getFeature(Feature.PERSON))
            frontVG.setFeature(Feature.NUMBER, numToUse)
            #don't push the front VG if it's a negated interrogative WH object question
            if not (phrase.getFeatureAsBoolean(Feature.NEGATED) and (InterrogativeType.WHO_OBJECT==interrogType or \
                    InterrogativeType.WHAT_OBJECT==interrogType)):
                vgComponents.append(frontVG)

    # Adds not to the stack if the phrase is negated.
    @classmethod
    def createNot(cls, phrase,  vgComponents, frontVG, hasModal):
        newFront = frontVG
        if phrase.getFeatureAsBoolean(Feature.NEGATED):
            factory = phrase.getFactory()
            # before adding "do", check if this is an object WH interrogative
            # in which case, don't add anything as it's already done by ClauseHelper
            interrType = phrase.getFeature(Feature.INTERROGATIVE_TYPE)
            addDo = not (InterrogativeType.WHAT_OBJECT==interrType or InterrogativeType.WHO_OBJECT==interrType)
            if vgComponents or frontVG is not None and cls.isCopular(frontVG):
                vgComponents.append(InflectedWordElement("not", LexicalCategory.ADVERB))
            else:
                if frontVG is not None and not hasModal:
                    frontVG.setFeature(Feature.NEGATED, True)
                    vgComponents.append(frontVG)
                vgComponents.append(InflectedWordElement("not", LexicalCategory.ADVERB))
                if addDo:
                    if factory is not None:
                        newFront = factory.createInflectedWord("do", LexicalCategory.VERB)
                    else:
                        newFront = InflectedWordElement("do", LexicalCategory.VERB)
        return newFront

    # Pushes the front verb on to the stack if the phrase has a modal.
    @classmethod
    def pushIfModal(cls, hasModal, phrase, frontVG, vgComponents):
        newFront = frontVG
        if hasModal and not phrase.getFeatureAsBoolean(InternalFeature.IGNORE_MODAL):
            if frontVG is not None:
                frontVG.setFeature(InternalFeature.NON_MORPH, True)
                vgComponents.append(frontVG)
            newFront = None
        return newFront

    # Adds have to the stack.
    @classmethod
    def addHave(cls, frontVG, vgComponents, modal, tenseValue):
        newFront = frontVG
        if frontVG is not None:
            frontVG.setFeature(Feature.FORM, Form.PAST_PARTICIPLE)
            vgComponents.append(frontVG)
        newFront = InflectedWordElement("have", LexicalCategory.VERB)
        newFront.setFeature(Feature.TENSE, tenseValue)
        if modal is not None:
            newFront.setFeature(InternalFeature.NON_MORPH, True)
        return newFront

    # Adds the be verb to the front of the group.
    @classmethod
    def addBe(cls, frontVG, vgComponents, frontForm):
        if frontVG is not None:
            frontVG.setFeature(Feature.FORM, frontForm)
            vgComponents.append(frontVG)
        return InflectedWordElement("be", LexicalCategory.VERB)

    # Checks to see if the phrase is in imperative, infinitive or bare
    # infinitive form. If it is then no morphology is done on the main verb.
    @classmethod
    def checkImperativeInfinitive(cls, formValue, frontVG):
        if (Form.IMPERATIVE==formValue or Form.INFINITIVE==formValue or Form.BARE_INFINITIVE==formValue) and \
                frontVG is not None:
            frontVG.setFeature(InternalFeature.NON_MORPH, True)

    # Grabs the head verb of the verb phrase and sets it to future tense if the
    # phrase is future tense. It also turns off negation if the group has a
    # modal.
    @classmethod
    def grabHeadVerb(cls, phrase,  tenseValue, hasModal):
        frontVG = phrase.getHead()
        if frontVG is not None:
            if isinstance(frontVG, WordElement):
                frontVG = InflectedWordElement(frontVG)
            # AG: tense value should always be set on frontVG
            if tenseValue is not None:
                frontVG.setFeature(Feature.TENSE, tenseValue)
            if hasModal:
                frontVG.setFeature(Feature.NEGATED, False)
        return frontVG

    # Pushes the particles of the main verb onto the verb group stack.
    @classmethod
    def pushParticles(cls, phrase, parent, vgComponents):
        particle = phrase.getFeature(Feature.PARTICLE)
        if isinstance(particle, str):
            vgComponents.append(StringElement(particle))
        elif isinstance(particle, NLGElement):
            vgComponents.append(parent.realise(particle))

    # Determines the number agreement for the phrase ensuring that any number
    # agreement on the parent element is inherited by the phrase.
    @classmethod
    def determineNumber(cls, parent,  phrase):
        numberValue = phrase.getFeature(Feature.NUMBER)
        if isinstance(numberValue, NumberAgreement):
            number = numberValue
        else:
            number = NumberAgreement.SINGULAR
        # Ehud Reiter = modified below to force number from VP for WHAT_SUBJECT
        # and WHO_SUBJECT interrogatuves
        if isinstance(parent, PhraseElement):
            if parent.isA(PhraseCategory.CLAUSE) and (PhraseHelper.isExpletiveSubject(parent) or \
                    InterrogativeType.WHO_SUBJECT==parent.getFeature(Feature.INTERROGATIVE_TYPE) or \
                    InterrogativeType.WHAT_SUBJECT==parent.getFeature(Feature.INTERROGATIVE_TYPE)) and \
                    cls.isCopular(phrase.getHead()):
                if cls.hasPluralComplement(phrase.getFeatureAsElementList(InternalFeature.COMPLEMENTS)):
                    number = NumberAgreement.PLURAL
                else:
                    number = NumberAgreement.SINGULAR
        return number

    # Checks to see if any of the complements to the phrase are plural.
    @classmethod
    def hasPluralComplement(cls, complements):
        plural = False
        for eachComplement in complements:
            if eachComplement is not None and eachComplement.isA(PhraseCategory.NOUN_PHRASE):
                numberValue = eachComplement.getFeature(Feature.NUMBER)
                if numberValue is not None and NumberAgreement.PLURAL==numberValue:
                    plural = True
                    break
        return plural

    # Checks to see if the base form of the word is copular, i.e. be.
    @classmethod
    def isCopular(cls, element):
        copular = False
        if isinstance(element, InflectedWordElement) or isinstance(element, WordElement):
            copular = "be" == element.getBaseForm().lower()
        elif isinstance(element, PhraseElement):
            # get the head and check if it's "be"
            if isinstance(element, SPhraseSpec):
                head = element.getVerb()
            else:
                head = element.getHead()
            if head is not None:
                copular = isinstance(head, WordElement) and "be"==head.getBaseForm()
        return copular



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# This enumerated type defines the different types of components found
# in the structure of text.
class DocumentCategory(Enum):
    # Definition for a document.
    DOCUMENT = 0
    # Definition for a section within a document.
    SECTION = 1
    # Definition for a paragraph.
    PARAGRAPH = 2
    # Definition for a sentence.
    SENTENCE = 3
    # Definition for creating a list of items.
    LIST = 4
    # Definition for creating a list of enumerated items.
    ENUMERATED_LIST = 5
    # Definition for an item in a list.
    LIST_ITEM = 6

    # Checks to see if the given object is equal to this document category
    def equalTo(self, checkObject):
        match = False
        if checkObject is not None:
            if isinstance(checkObject, DocumentCategory):
                match = self==checkObject
            else:
                match = str(self.name).lower() == str(checkObject).lower()
        return match

    # This method determines if the supplied ElementCategory (ecat) forms an immediate
    # sub-part of DocumentCategory (dcat)
    def hasSubPart(self, ecat):
        if isinstance(ecat, DocumentCategory):
            if self == DocumentCategory.DOCUMENT:
                return not (ecat==DocumentCategory.DOCUMENT) and not (ecat==DocumentCategory.LIST_ITEM)
            elif self == DocumentCategory.SECTION:
                return ecat==DocumentCategory.PARAGRAPH or ecat==DocumentCategory.SECTION
            elif self == DocumentCategory.PARAGRAPH:
                return ecat==DocumentCategory.SENTENCE or ecat==DocumentCategory.LIST
            elif self == DocumentCategory.LIST:
                return ecat==DocumentCategory.LIST_ITEM
            elif self == DocumentCategory.ENUMERATED_LIST:
                return ecat==DocumentCategory.LIST_ITEM
        else:
            return self==DocumentCategory.SENTENCE or self==DocumentCategory.LIST_ITEM
        return False



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from enum import Enum

# This enumeration defines the different lexical components.
class LexicalCategory(Enum):
    ANY             =  0    # A default value, indicating an unspecified category.
    SYMBOL          =  1    # The element represents a symbol.
    NOUN            =  2    # A noun element.
    ADJECTIVE       =  3    # An adjective element.
    ADVERB          =  4    # An adverb element.
    VERB            =  5    # A verb element.
    DETERMINER      =  6    # A determiner element often referred to as a specifier.
    PRONOUN         =  7    # A pronoun element.
    CONJUNCTION     =  8    # A conjunction element.
    PREPOSITION     =  9    # A preposition element.
    COMPLEMENTISER  = 10    # A complementiser element.
    MODAL           = 11    #  A modal element.
    AUXILIARY       = 12    # An auxiliary verb element.

    # Checks to see if the given object is equal to this document category
    def equalTo(self, checkObject):
        match = False
        if checkObject is not None:
            if isinstance(checkObject, LexicalCategory):
                match = self==checkObject
            else:
                match = str(self.name).lower() == str(checkObject).lower()
        return match

    def __str__(self):
        return self.name



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.


# This class defines a list of constant values used by SimpleNLG lexicons. 
class LexicalFeature(object):
    # This feature is used to map an acronym element to the full forms of the
    # acronym.
    ACRONYM_OF = "acronym_of"
    # This feature is used to map a word to its acronyms.
    ACRONYMS = "acronyms"
    # This feature is used to list all the possible inflectional variants
    # of a word.
    INFLECTIONS = "infl"    # note - commented out in original code
    # This feature is used to specify, for a given word, what its default
    # inflectional variant is, if more than one is possible.
    DEFAULT_INFL = "default_infl"
    # This feature is used to specify the spelling variants of a word.
    SPELL_VARS = "spell_vars"
    # This feature is used to specify the default spelling variant of a word,
    # if it has more than one.
    DEFAULT_SPELL = "default_spell"
    # This feature is used to define the base form for phrases and words.
    BASE_FORM = "base_form"
    # This feature is used for determining the position of adjectives. Setting
    # this value to true means that the adjective can occupy the
    CLASSIFYING = "classifying"
    # This feature is used for determining the position of adjectives. Setting
    # this value to true means that the adjective can occupy the
    # colour position.
    COLOUR = "colour"
    # This feature gives the comparative form for adjectives and adverbs. For
    # example, dizzier is the comparative form of dizzy,
    # fatter is the comparative form of fat and
    # earlier is the comparative form of early.
    COMPARATIVE = "comparative"
    # This feature determines if a verb is ditransitive, meaning that it can
    # have a subject, direct object and indirect object. For example in the
    # phrase he gave Mary ten pounds, the verb give has three
    # components: the subject is the person doing the giving (he), the
    # direct object is the object being passed (ten pounds) and the
    # indirect object is the recipient (Mary).
    DITRANSITIVE = "ditransitive"
    # This feature determines whether a noun is masculine, feminine or neuter
    # in nature.
    GENDER = "gender"
    # This flag determines if an adverb is an intensifier, such as
    # very.
    INTENSIFIER = "intensifier"
    # This flag highlights a verb that can only take a subject and no objects.
    INTRANSITIVE = "intransitive"
    # <p> This feature represents non-countable nouns such as mud,
    # sand and water.
    NON_COUNT = "nonCount"     # commented out in original code
    # This feature gives the past tense form of a verb. For example, the past
    # tense of eat is ate, the past tense of walk is
    # walked.
    PAST = "past"
    # This feature gives the past participle tense form of a verb. For many
    # verbs the past participle is exactly the same as the past tense
    PAST_PARTICIPLE = "pastParticiple"
    # This feature gives the plural form of a noun.
    PLURAL = "plural"
    # This flag is set on adjectives that can also be used as a predicate.
    PREDICATIVE = "predicative"
    # This feature gives the present participle form of a verb.
    PRESENT_PARTICIPLE = "presentParticiple"
    # This feature gives the present third person singular form of a verb.
    PRESENT3S = "present3s"
    # This flag is used to determine whether a noun is a proper noun, such as a
    # person's name.
    PROPER = "proper"
    # This feature is used for determining the position of adjectives. Setting
    # this value to true means that the adjective can occupy the
    # qualitative position.
    QUALITATIVE = "qualitative"
    # This flag is set if a pronoun is written in the reflexive form.
    REFLEXIVE = "reflexive"
    # This feature is used to define whether an adverb can be used as a clause
    # modifier, which are normally applied at the beginning of clauses.
    SENTENCE_MODIFIER = "sentence_modifier"
    # This feature gives the superlative form for adjectives and adverbs.
    SUPERLATIVE = "superlative"
    # This flag highlights a verb that can only take a subject and an object.
    TRANSITIVE = "transitive"
    # This feature is used to define whether an adverb can be used as a verb
    # modifier, which are normally added in a phrase before the verb itself.
    VERB_MODIFIER = "verb_modifier"
    # This feature determines if the pronoun is an expletive or not. Expletive
    # pronouns are usually it or there in sentences such as:
    # It is raining now.
    # There are ten desks in the room.
    EXPLETIVE_SUBJECT = "expletive_subject"

    # Return those features related to a word's inflection, depending on its
    # category, that is, the constants for
    # PAST, PAST_PARTICIPLE, PLURAl, PRESENT_PARTICIPLE, PRESENT3S, COMPARATIVE
    # or SUPERLATIVE.
    @classmethod
    def getInflectionalFeatures(cls, element_category):
        from ..framework.LexicalCategory  import LexicalCategory    # prevent circular import
        from ..framework.PhraseCategory   import PhraseCategory
        if PhraseCategory.NOUN_PHRASE==element_category or LexicalCategory.NOUN==element_category:
            return [cls.PLURAL]
        elif PhraseCategory.VERB_PHRASE==element_category or LexicalCategory.VERB==element_category:
            return [cls.PAST, cls.PAST_PARTICIPLE, cls.PRESENT_PARTICIPLE, cls.PRESENT3S]
        elif PhraseCategory.ADJECTIVE_PHRASE==element_category or LexicalCategory.ADJECTIVE==element_category:
            return [cls.COMPARATIVE, cls.SUPERLATIVE]
        else:
            return None



from . CoordinatedPhraseElement import *
from . DocumentElement          import *
from . DocumentCategory         import *
from . InflectedWordElement     import *
from . LexicalCategory          import *
from . ListElement              import *
from . NLGElement               import *
from . NLGFactory               import *
from . NLGModule                import *
from . PhraseElement            import *
from . PhraseCategory           import *
from . StringElement            import *
from . WordElement              import *



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..framework.LexicalCategory    import *
from ..framework.NLGElement         import *
from ..framework.PhraseCategory     import *
from ..framework.PhraseElement      import *


# This class defines a adverbial phrase.  
class AdvPhraseSpec(PhraseElement):
    def __init__(self,  phraseFactory):
        super().__init__(PhraseCategory.ADVERB_PHRASE)
        self.setFactory(phraseFactory)

    # sets the adverb (head) of the phrase
    def setAdverb(self, adverb):
        if isinstance(adverb, NLGElement):
            self.setHead(adverb)
        else :
            adverbElement = self.getFactory().createWord(adverb, LexicalCategory.ADVERB)
            self.setHead(adverbElement)

    # @return adverb (head) of  phrase
    def getAdverb(self):
        return self.getHead()



# The contents of this file are subject to the Mozilla Public License
# Version 1.1 (the "License"); you may not use this file except in
# compliance with the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS"
# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
# License for the specific language governing rights and limitations
# under the License.
#
# The Original Code is "Simplenlg".
#
# The Initial Developer of the Original Code is Ehud Reiter, Albert Gatt and Dave Westwater.
# Portions created by Ehud Reiter, Albert Gatt and Dave Westwater are
# Copyright (C) 2010-11 The University of Aberdeen. All Rights Reserved.
#
# Contributor(s): Ehud Reiter, Albert Gatt, Dave Wewstwater, Roman Kutlak, Margaret Mitchell.

from ..framework.NLGElement     import *
from ..framework.PhraseCategory import *
from ..framework.PhraseElement  import *


# This class defines a adjective phrase.  
class AdjPhraseSpec(PhraseElement):
    def __init__(self, phraseFactory):
        super().__init__(PhraseCategory.ADJECTIVE_PHRASE)
        self.setFactory(phraseFactory)

    # sets the adjective (head) of the phrase
    def setAdjective(self, adjective):
        if isinstance(adjective, NLGElement):
            self.setHead(adjective)
        else:
            adjectiveElement = self.getFactory().createWord(adjective, LexicalCategory.ADJECTIVE)
            self.setHead(adjectiveElement)

    # @return adjective (head) of  phrase
    def getAdjective(self):
        return self.getHead()
